# sage_setup: distribution = sagemath-schemes
r"""
Jacobian 'morphism' as a class in the Picard group

This module implements the group operation in the Picard group of a
hyperelliptic curve, represented as divisors in Mumford
representation, using Cantor's algorithm.

A divisor on the hyperelliptic curve `y^2 + y h(x) = f(x)`
is stored in Mumford representation, that is, as two polynomials
`u(x)` and `v(x)` such that:

- `u(x)` is monic,

- `u(x)` divides `f(x) - h(x) v(x) - v(x)^2`,

- `deg(v(x)) < deg(u(x)) \le g`.

REFERENCES:

A readable introduction to divisors, the Picard group, Mumford
representation, and Cantor's algorithm:

- J. Scholten, F. Vercauteren. An Introduction to Elliptic and
  Hyperelliptic Curve Cryptography and the NTRU Cryptosystem. To
  appear in B. Preneel (Ed.) State of the Art in Applied Cryptography
  - COSIC '03, Lecture Notes in Computer Science, Springer 2004.

A standard reference in the field of cryptography:

- R. Avanzi, H. Cohen, C. Doche, G. Frey, T. Lange, K. Nguyen, and F.
  Vercauteren, Handbook of Elliptic and Hyperelliptic Curve
  Cryptography. CRC Press, 2005.

EXAMPLES: The following curve is the reduction of a curve whose
Jacobian has complex multiplication.

::

    sage: x = GF(37)['x'].gen()
    sage: H = HyperellipticCurve(x^5 + 12*x^4 + 13*x^3 + 15*x^2 + 33*x); H
    Hyperelliptic Curve over Finite Field of size 37 defined
     by y^2 = x^5 + 12*x^4 + 13*x^3 + 15*x^2 + 33*x

At this time, Jacobians of hyperelliptic curves are handled
differently than elliptic curves::

    sage: J = H.jacobian(); J
    Jacobian of Hyperelliptic Curve over Finite Field of size 37 defined
     by y^2 = x^5 + 12*x^4 + 13*x^3 + 15*x^2 + 33*x
    sage: J = J(J.base_ring()); J
    Set of rational points of Jacobian of Hyperelliptic Curve over Finite Field
     of size 37 defined by y^2 = x^5 + 12*x^4 + 13*x^3 + 15*x^2 + 33*x

Points on the Jacobian are represented by Mumford's polynomials.
First we find a couple of points on the curve::

    sage: P1 = H.lift_x(2); P1
    (2 : 11 : 1)
    sage: Q1 = H.lift_x(10); Q1
    (10 : 18 : 1)

Observe that 2 and 10 are the roots of the polynomials in x,
respectively::

    sage: P = J(P1); P
    (x + 35, y + 26)
    sage: Q = J(Q1); Q
    (x + 27, y + 19)

::

    sage: P + Q
    (x^2 + 25*x + 20, y + 13*x)
    sage: (x^2 + 25*x + 20).roots(multiplicities=False)
    [10, 2]

Frobenius satisfies

.. MATH::

    x^4 + 12*x^3 + 78*x^2 + 444*x + 1369

on the Jacobian of this reduction and the order of the Jacobian is
`N = 1904`.

::

    sage: 1904*P
    (1)
    sage: 34*P == 0
    True
    sage: 35*P == P
    True
    sage: 33*P == -P
    True

::

    sage: Q*1904
    (1)
    sage: Q*238 == 0
    True
    sage: Q*239 == Q
    True
    sage: Q*237 == -Q
    True
"""

#*****************************************************************************
#  Copyright (C) 2005 David Kohel <kohel@maths.usyd.edu.au>
#  Distributed under the terms of the GNU General Public License (GPL)
#                  http://www.gnu.org/licenses/
#*****************************************************************************

from sage.misc.latex import latex

from sage.structure.element import AdditiveGroupElement
from sage.structure.richcmp import richcmp, op_NE
from sage.schemes.generic.morphism import SchemeMorphism


def cantor_reduction_simple(a, b, f, genus):
    r"""
    Return the unique reduced divisor linearly equivalent to
    `(a, b)` on the curve `y^2 = f(x).`

    See the docstring of
    :mod:`sage.schemes.hyperelliptic_curves.jacobian_morphism` for
    information about divisors, linear equivalence, and reduction.

    EXAMPLES::

        sage: x = QQ['x'].gen()
        sage: f = x^5 - x
        sage: H = HyperellipticCurve(f); H
        Hyperelliptic Curve over Rational Field defined by y^2 = x^5 - x
        sage: J = H.jacobian()(QQ); J
        Set of rational points of Jacobian of Hyperelliptic Curve over Rational Field
         defined by y^2 = x^5 - x

    The following point is 2-torsion::

        sage: P = J(H.lift_x(-1)); P
        (x + 1, y)
        sage: 2 * P # indirect doctest
        (1)
    """
    a2 = (f - b**2) // a
    a2 = a2.monic()
    b2 = -b % (a2)
    if a2.degree() == a.degree():
        # XXX
        assert a2.degree() == genus + 1
        print("Returning ambiguous form of degree genus+1.")
        return (a2, b2)
    elif a2.degree() > genus:
        return cantor_reduction_simple(a2, b2, f, genus)
    return (a2, b2)


def cantor_reduction(a, b, f, h, genus):
    r"""
    Return the unique reduced divisor linearly equivalent to
    `(a, b)` on the curve `y^2 + y h(x) = f(x)`.

    See the docstring of
    :mod:`sage.schemes.hyperelliptic_curves.jacobian_morphism` for
    information about divisors, linear equivalence, and reduction.

    EXAMPLES::

        sage: x = QQ['x'].gen()
        sage: f = x^5 - x
        sage: H = HyperellipticCurve(f, x); H
        Hyperelliptic Curve over Rational Field defined by y^2 + x*y = x^5 - x
        sage: J = H.jacobian()(QQ); J
        Set of rational points of Jacobian of Hyperelliptic Curve over
         Rational Field defined by y^2 + x*y = x^5 - x

    The following point is 2-torsion::

        sage: Q = J(H.lift_x(0)); Q
        (x, y)
        sage: 2*Q # indirect doctest
        (1)

    The next point is not 2-torsion::

        sage: P = J(H.lift_x(-1)); P
        (x + 1, y)
        sage: 2 * J(H.lift_x(-1)) # indirect doctest
        (x^2 + 2*x + 1, y + 4*x + 4)
        sage: 3 * J(H.lift_x(-1)) # indirect doctest
        (x^2 - 487*x - 324, y + 10755*x + 7146)
    """
    assert a.degree() < 2*genus+1
    assert b.degree() < a.degree()
    k = f - h*b - b**2
    if 2*a.degree() == k.degree():
        # must adjust b to include the point at infinity
        g1 = a.degree()
        x = a.parent().gen()
        r = (x**2 + h[g1]*x - f[2*g1]).roots()[0][0]
        b = b + r*(x**g1 - (x**g1) % (a))
        k = f - h*b - b**2
    assert k % (a) == 0
    a = (k // a).monic()
    b = -(b+h) % (a)
    if a.degree() > genus:
        return cantor_reduction(a, b, f, h, genus)
    return (a, b)


def cantor_composition_simple(D1, D2, f, genus):
    r"""
    Given `D_1` and `D_2` two reduced Mumford
    divisors on the Jacobian of the curve `y^2 = f(x)`,
    computes a representative `D_1 + D_2`.

    .. warning::

       The representative computed is NOT reduced! Use
       :func:`cantor_reduction_simple` to reduce it.

    EXAMPLES::

        sage: x = QQ['x'].gen()
        sage: f = x^5 + x
        sage: H = HyperellipticCurve(f); H
        Hyperelliptic Curve over Rational Field defined by y^2 = x^5 + x

    ::

        sage: F.<a> = NumberField(x^2 - 2, 'a')                                         # needs sage.rings.number_field
        sage: J = H.jacobian()(F); J                                                    # needs sage.rings.number_field
        Set of rational points of Jacobian of Hyperelliptic Curve over
         Number Field in a with defining polynomial x^2 - 2 defined by y^2 = x^5 + x

    ::

        sage: # needs sage.rings.number_field
        sage: P = J(H.lift_x(F(1))); P
        (x - 1, y + a)
        sage: Q = J(H.lift_x(F(0))); Q
        (x, y)
        sage: 2*P + 2*Q # indirect doctest
        (x^2 - 2*x + 1, y + 3/2*a*x - 1/2*a)
        sage: 2*(P + Q) # indirect doctest
        (x^2 - 2*x + 1, y + 3/2*a*x - 1/2*a)
        sage: 3*P # indirect doctest
        (x^2 - 25/32*x + 49/32, y + 45/256*a*x + 315/256*a)
    """
    a1, b1 = D1
    a2, b2 = D2
    if a1 == a2 and b1 == b2:
        # Duplication law:
        d, h1, h3 = a1.xgcd(2*b1)
        a = (a1 // d)**2
        b = (b1 + h3*((f - b1**2) // d)) % (a)
    else:
        d0, _, h2 = a1.xgcd(a2)
        if d0 == 1:
            a = a1*a2
            b = (b2 + h2*a2*(b1-b2)) % (a)
        else:
            d, l, h3 = d0.xgcd(b1 + b2)
            a = (a1*a2) // (d**2)
            b = ((b2 + l*h2*(b1-b2)*(a2 // d)) + h3*((f - b2**2) // d)) % (a)
    a = a.monic()
    return (a, b)


def cantor_composition(D1, D2, f, h, genus):
    r"""
    EXAMPLES::

        sage: # needs sage.rings.finite_rings
        sage: F.<a> = GF(7^2, 'a')
        sage: x = F['x'].gen()
        sage: f = x^7 + x^2 + a
        sage: H = HyperellipticCurve(f, 2*x); H
        Hyperelliptic Curve over Finite Field in a of size 7^2
         defined by y^2 + 2*x*y = x^7 + x^2 + a
        sage: J = H.jacobian()(F); J
        Set of rational points of Jacobian of Hyperelliptic Curve over
         Finite Field in a of size 7^2 defined by y^2 + 2*x*y = x^7 + x^2 + a

    ::

        sage: Q = J(H.lift_x(F(1))); Q                                                  # needs sage.rings.finite_rings
        (x + 6, y + 5*a)
        sage: 10*Q  # indirect doctest                                                  # needs sage.rings.finite_rings
        (x^3 + (3*a + 1)*x^2 + (2*a + 5)*x + a + 5, y + (3*a + 2)*x^2 + (6*a + 1)*x + a + 4)
        sage: 7*8297*Q                                                                  # needs sage.rings.finite_rings
        (1)

    ::

        sage: Q = J(H.lift_x(F(a+1))); Q                                                # needs sage.rings.finite_rings
        (x + 6*a + 6, y + 2)
        sage: 7*8297*Q  # indirect doctest                                              # needs sage.rings.finite_rings
        (1)

        A test over a prime field:

        sage: # needs sage.rings.finite_rings
        sage: F = GF(next_prime(10^30))
        sage: x = F['x'].gen()
        sage: f = x^7 + x^2 + 1
        sage: H = HyperellipticCurve(f, 2*x); H
        Hyperelliptic Curve over Finite Field of size 1000000000000000000000000000057
         defined by y^2 + 2*x*y = x^7 + x^2 + 1
        sage: J = H.jacobian()(F); J
        Set of rational points of Jacobian of Hyperelliptic Curve
         over Finite Field of size 1000000000000000000000000000057
         defined by y^2 + 2*x*y = x^7 + x^2 + 1
        sage: Q = J(H.lift_x(F(1))); Q
        (x + 1000000000000000000000000000056, y + 1000000000000000000000000000056)
        sage: 10*Q  # indirect doctest
        (x^3 + 150296037169838934997145567227*x^2
             + 377701248971234560956743242408*x + 509456150352486043408603286615,
         y + 514451014495791237681619598519*x^2
           + 875375621665039398768235387900*x + 861429240012590886251910326876)
        sage: 7*8297*Q
        (x^3 + 35410976139548567549919839063*x^2
             + 26230404235226464545886889960*x + 681571430588959705539385624700,
         y + 999722365017286747841221441793*x^2
           + 262703715994522725686603955650*x + 626219823403254233972118260890)
    """
    a1, b1 = D1
    a2, b2 = D2
    if a1 == a2 and b1 == b2:
        # Duplication law:
        d, h1, h3 = a1.xgcd(2*b1 + h)
        a = (a1 // d)**2
        b = (b1 + h3*((f-h*b1-b1**2) // d)) % (a)
    else:
        d0, _, h2 = a1.xgcd(a2)
        if d0 == 1:
            a = a1 * a2
            b = (b2 + h2*a2*(b1-b2)) % (a)
        else:
            e0 = b1+b2+h
            if e0 == 0:
                a = (a1*a2) // (d0**2)
                b = (b2 + h2*(b1-b2)*(a2 // d0)) % (a)
            else:
                d, l, h3 = d0.xgcd(e0)
                a = (a1*a2) // (d**2)
                b = (b2 + l*h2*(b1-b2)*(a2 // d) + h3*((f-h*b2-b2**2) // d)) % (a)
    a = a.monic()
    return (a, b)


class JacobianMorphism_divisor_class_field(AdditiveGroupElement, SchemeMorphism):
    r"""
    An element of a Jacobian defined over a field, i.e. in
    `J(K) = \mathrm{Pic}^0_K(C)`.
    """
    def __init__(self, parent, polys, check=True):
        r"""
        Create a new Jacobian element in Mumford representation.

        INPUT:

        - ``parent`` -- the parent Homset
        - ``polys`` -- Mumford's `u` and `v` polynomials
        - ``check`` -- boolean (default: ``True``); if ``True``, ensure that
          polynomials define a divisor on the appropriate curve and are reduced

        .. warning::

           Not for external use! Use ``J(K)([u, v])`` instead.

        EXAMPLES::

            sage: x = GF(37)['x'].gen()
            sage: H = HyperellipticCurve(x^5 + 12*x^4 + 13*x^3 + 15*x^2 + 33*x)
            sage: J = H.jacobian()(GF(37));  J
            Set of rational points of Jacobian of Hyperelliptic Curve over
             Finite Field of size 37 defined by
            y^2 = x^5 + 12*x^4 + 13*x^3 + 15*x^2 + 33*x

        ::

            sage: P1 = J(H.lift_x(2)); P1  # indirect doctest
            (x + 35, y + 26)
            sage: P1.parent()
            Set of rational points of Jacobian of Hyperelliptic Curve over
             Finite Field of size 37 defined by
            y^2 = x^5 + 12*x^4 + 13*x^3 + 15*x^2 + 33*x
            sage: type(P1)
            <class 'sage.schemes.hyperelliptic_curves.jacobian_morphism.JacobianMorphism_divisor_class_field'>
        """
        SchemeMorphism.__init__(self, parent)
        if check:
            C = parent.curve()
            f, h = C.hyperelliptic_polynomials()
            a, b = polys
            if not (b**2 + h*b - f) % a == 0:
                raise ValueError("Argument polys (= %s) must be divisor on curve %s." % (
                    polys, C))
            genus = C.genus()
            if a.degree() > genus:
                polys = cantor_reduction(a, b, f, h, genus)
        self.__polys = tuple(polys)

    def _printing_polys(self):
        r"""
        Internal function formatting Mumford polynomials for printing.

        TESTS::

            sage: # needs sage.rings.finite_rings
            sage: F.<a> = GF(7^2, 'a')
            sage: x = F['x'].gen()
            sage: f = x^7 + x^2 + a
            sage: H = HyperellipticCurve(f, 2*x)
            sage: J = H.jacobian()(F)

        ::

            sage: Q = J(H.lift_x(F(1))); Q  # indirect doctest                          # needs sage.rings.finite_rings
            (x + 6, y + 5*a)
        """
        a, b = self.__polys
        P = self.parent()._printing_ring
        y = P.gen()
        x = P.base_ring().gen()
        return (a(x), y - b(x))

    def _repr_(self):
        r"""
        Return a string representation of this Mumford divisor.

        EXAMPLES::

            sage: # needs sage.rings.finite_rings
            sage: F.<a> = GF(7^2, 'a')
            sage: x = F['x'].gen()
            sage: f = x^7 + x^2 + a
            sage: H = HyperellipticCurve(f, 2*x)
            sage: J = H.jacobian()(F)

        ::

            sage: Q = J(0); Q  # indirect doctest                                       # needs sage.rings.finite_rings
            (1)
            sage: Q = J(H.lift_x(F(1))); Q  # indirect doctest                          # needs sage.rings.finite_rings
            (x + 6, y + 5*a)
            sage: Q + Q  # indirect doctest                                             # needs sage.rings.finite_rings
            (x^2 + 5*x + 1, y + (4*a + 2)*x + a + 5)
        """
        if self.is_zero():
            return "(1)"
        a, b = self._printing_polys()
        return "(%s, %s)" % (a, b)

    def _latex_(self):
        r"""
        Return a LaTeX string representing this Mumford divisor.

        EXAMPLES::

            sage: # needs sage.rings.finite_rings
            sage: F.<alpha> = GF(7^2)
            sage: x = F['x'].gen()
            sage: f = x^7 + x^2 + alpha
            sage: H = HyperellipticCurve(f, 2*x)
            sage: J = H.jacobian()(F)

        ::

            sage: Q = J(0); print(latex(Q))  # indirect doctest                         # needs sage.rings.finite_rings
            \left(1\right)
            sage: Q = J(H.lift_x(F(1))); print(latex(Q))  # indirect doctest            # needs sage.rings.finite_rings
            \left(x + 6, y + 5 \alpha\right)

        ::

            sage: print(latex(Q + Q))                                                   # needs sage.rings.finite_rings
            \left(x^{2} + 5 x + 1, y + \left(4 \alpha + 2\right) x + \alpha + 5\right)
        """
        if self.is_zero():
            return "\\left(1\\right)"
        a, b = self._printing_polys()
        return "\\left(%s, %s\\right)" % (latex(a), latex(b))

    def scheme(self):
        r"""
        Return the scheme this morphism maps to; or, where this divisor lives.

        .. WARNING::

            Although a pointset is defined over a specific field, the
            scheme returned may be over a different (usually smaller)
            field.  The example below demonstrates this: the pointset
            is determined over a number field of absolute degree 2 but
            the scheme returned is defined over the rationals.

        EXAMPLES::

            sage: # needs sage.rings.number_field
            sage: x = QQ['x'].gen()
            sage: f = x^5 + x
            sage: H = HyperellipticCurve(f)
            sage: F.<a> = NumberField(x^2 - 2, 'a')
            sage: J = H.jacobian()(F); J
            Set of rational points of Jacobian of Hyperelliptic Curve
             over Number Field in a with defining polynomial x^2 - 2
             defined by y^2 = x^5 + x
            sage: P = J(H.lift_x(F(1)))
            sage: P.scheme()
            Jacobian of Hyperelliptic Curve over Rational Field defined by y^2 = x^5 + x
        """
        return self.codomain()

    def point_of_jacobian_of_curve(self):
        r"""
        Return the point in the Jacobian of the curve.

        The Jacobian is the one attached to the projective curve
        corresponding to this hyperelliptic curve.

        EXAMPLES::

            sage: R.<x> = PolynomialRing(GF(11))
            sage: f = x^6 + x + 1
            sage: H = HyperellipticCurve(f)
            sage: J = H.jacobian()
            sage: D = J(H.lift_x(1))
            sage: D  # divisor in Mumford representation
            (x + 10, y + 6)
            sage: jacobian_order = sum(H.frobenius_polynomial())
            sage: jacobian_order
            234
            sage: p = D.point_of_jacobian_of_curve(); p
            [Place (1/x0, 1/x0^3*x1 + 1)
             + Place (x0 + 10, x1 + 6)]
            sage: p  # Jacobian point represented by an effective divisor
            [Place (1/x0, 1/x0^3*x1 + 1)
             + Place (x0 + 10, x1 + 6)]
            sage: p.order()
            39
            sage: 234*p == 0
            True
            sage: G = p.parent()
            sage: G
            Group of rational points of Jacobian over Finite Field of size 11 (Hess model)
            sage: J = G.parent()
            sage: J
            Jacobian of Projective Plane Curve over Finite Field of size 11
             defined by x0^6 + x0^5*x1 + x1^6 - x0^4*x2^2 (Hess model)
            sage: C = J.curve()
            sage: C
            Projective Plane Curve over Finite Field of size 11
             defined by x0^6 + x0^5*x1 + x1^6 - x0^4*x2^2
            sage: C.affine_patch(0) == H.affine_patch(2)
            True
        """
        from sage.schemes.curves.constructor import Curve
        C = self.parent().curve()
        P = C.ambient_space()  # projective plane
        x0, x1, x2 = P.gens()

        # X is the curve positioned in the ambient space
        # such that x1 = x and x2 = y
        X = Curve(C.defining_ideal().gens(), P)
        X = X.affine_patch(2).projective_closure()

        u0, v0 = list(self)
        u1 = u0.subs(x1).homogenize(x0)
        v1 = (x2 - v0.subs(x1)).homogenize(x0)
        u2 = u1/x0**u1.degree()
        v2 = v1/x0**v1.degree()
        u = X.function(u2)
        v = X.function(v2)

        F = X.function_field()
        O = F.maximal_order()
        D = O.ideal([u,v]).divisor()

        Pinf = F.places_infinite()[0]
        assert Pinf.degree() == 1, "no rational point at infinity"

        J = X.jacobian(model='hess', base_div=F.genus()*Pinf)
        G = J.group(self.base_ring())
        return G(D - D.degree()*Pinf)

    def __list__(self):
        r"""
        Return a list `(a(x), b(x))` of the polynomials giving the
        Mumford representation of ``self``.

        TESTS::

            sage: x = QQ['x'].gen()
            sage: f = x^5 + x
            sage: H = HyperellipticCurve(f)
            sage: F.<a> = NumberField(x^2 - 2, 'a')                                     # needs sage.rings.number_field
            sage: J = H.jacobian()(F); J                                                # needs sage.rings.number_field
            Set of rational points of Jacobian of Hyperelliptic Curve
             over Number Field in a with defining polynomial x^2 - 2
             defined by y^2 = x^5 + x

        ::

            sage: P = J(H.lift_x(F(1)))                                                 # needs sage.rings.number_field
            sage: list(P)  # indirect doctest                                           # needs sage.rings.number_field
            [x - 1, -a]
        """
        return list(self.__polys)

    def __tuple__(self):
        r"""
        Return a tuple `(a(x), b(x))` of the polynomials giving the
        Mumford representation of ``self``.

        TESTS::

            sage: x = QQ['x'].gen()
            sage: f = x^5 + x
            sage: H = HyperellipticCurve(f)
            sage: F.<a> = NumberField(x^2 - 2, 'a')                                     # needs sage.rings.number_field
            sage: J = H.jacobian()(F); J                                                # needs sage.rings.number_field
            Set of rational points of Jacobian of Hyperelliptic Curve
             over Number Field in a with defining polynomial x^2 - 2
             defined by y^2 = x^5 + x

        ::

            sage: P = J(H.lift_x(F(1)))                                                 # needs sage.rings.number_field
            sage: tuple(P)  # indirect doctest                                          # needs sage.rings.number_field
            (x - 1, -a)
        """
        return self.__polys

    def __getitem__(self, n):
        r"""
        Return the `n`-th item of the pair `(a(x), b(x))`
        of polynomials giving the Mumford representation of ``self``.

        TESTS::

            sage: x = QQ['x'].gen()
            sage: f = x^5 + x
            sage: H = HyperellipticCurve(f)
            sage: F.<a> = NumberField(x^2 - 2, 'a')                                     # needs sage.rings.number_field
            sage: J = H.jacobian()(F); J                                                # needs sage.rings.number_field
            Set of rational points of Jacobian of Hyperelliptic Curve
             over Number Field in a with defining polynomial x^2 - 2
             defined by y^2 = x^5 + x

        ::

            sage: # needs sage.rings.number_field
            sage: P = J(H.lift_x(F(1)))
            sage: P[0] # indirect doctest
            x - 1
            sage: P[1] # indirect doctest
            -a
            sage: P[-1] # indirect doctest
            -a
            sage: P[:1] # indirect doctest
            (x - 1,)
        """
        return self.__polys[n]

    def _richcmp_(self, other, op):
        r"""
        Compare ``self`` and ``other``.

        TESTS::

            sage: x = QQ['x'].gen()
            sage: f = x^5 - x
            sage: H = HyperellipticCurve(f); H
            Hyperelliptic Curve over Rational Field defined by y^2 = x^5 - x
            sage: J = H.jacobian()(QQ); J
            Set of rational points of Jacobian of Hyperelliptic Curve over
            Rational Field defined by y^2 = x^5 - x

        The following point is 2-torsion::

            sage: P = J(H.lift_x(-1)); P
            (x + 1, y)
            sage: 0 == 2 * P # indirect doctest
            True
            sage: P == P
            True

        ::

            sage: Q = J(H.lift_x(-1))
            sage: Q == P
            True

        ::

            sage: 2 == Q
            False
            sage: P == False
            False

        Let's verify the same "points" on different schemes are not equal::

            sage: x = QQ['x'].gen()
            sage: f = x^5 + x
            sage: H2 = HyperellipticCurve(f)
            sage: J2 = H2.jacobian()(QQ)

        ::

            sage: P1 = J(H.lift_x(0)); P1
            (x, y)
            sage: P2 = J2(H2.lift_x(0)); P2
            (x, y)
            sage: P1 == P2
            False
        """
        if self.scheme() != other.scheme():
            return op == op_NE
        # since divisors are internally represented as Mumford divisors,
        # comparing polynomials is well-defined
        return richcmp(self.__polys, other.__polys, op)

    def __hash__(self):
        return hash(self.__polys)

    def __bool__(self):
        r"""
        Return ``True`` if this divisor is not the additive identity element.

        EXAMPLES::

            sage: x = GF(37)['x'].gen()
            sage: H = HyperellipticCurve(x^5 + 12*x^4 + 13*x^3 + 15*x^2 + 33*x)
            sage: J = H.jacobian()(GF(37))

        ::

            sage: P1 = J(H.lift_x(2)); P1
            (x + 35, y + 26)
            sage: P1 == 0  # indirect doctest
            False
            sage: P1 - P1 == 0  # indirect doctest
            True
        """
        return self.__polys[0] != 1

    def __neg__(self):
        r"""
        Return the additive inverse of this divisor.

        EXAMPLES::

            sage: x = GF(37)['x'].gen()
            sage: H = HyperellipticCurve(x^5 + 12*x^4 + 13*x^3 + 15*x^2 + 33*x)
            sage: J = H.jacobian()(GF(37))
            sage: P1 = J(H.lift_x(2)); P1
            (x + 35, y + 26)
            sage: - P1  # indirect doctest
            (x + 35, y + 11)
            sage: P1 + (-P1)  # indirect doctest
            (1)

        ::

            sage: H2 = HyperellipticCurve(x^5 + 12*x^4 + 13*x^3 + 15*x^2 + 33*x, x)
            sage: J2 = H2.jacobian()(GF(37))
            sage: P2 = J2(H2.lift_x(2)); P2
            (x + 35, y + 24)
            sage: - P2  # indirect doctest
            (x + 35, y + 15)
            sage: P2 + (-P2)  # indirect doctest
            (1)

        TESTS:

        The following was fixed in :issue:`14264`::

            sage: # needs sage.rings.number_field
            sage: P.<x> = QQ[]
            sage: f = x^5 - x + 1; h = x
            sage: C = HyperellipticCurve(f, h, 'u,v')
            sage: J = C.jacobian()
            sage: K.<t> = NumberField(x^2 - 2)
            sage: R.<x> = K[]
            sage: Q = J(K)([x^2 - t, R(1)])
            sage: Q
            (u^2 - t, v - 1)
            sage: -Q
            (u^2 - t, v + u + 1)
            sage: Q + (-Q)  # indirect doctest
            (1)
        """
        if self.is_zero():
            return self
        polys = self.__polys
        X = self.parent()
        f, h = X.curve().hyperelliptic_polynomials()
        if h.is_zero():
            D = (polys[0],-polys[1])
        else:
            # It is essential that the modulus polys[0] can be converted into
            # the parent of the dividend h. This is not always automatically
            # the case (h can be a rational polynomial and polys[0] can a
            # non-constant polynomial over a number field different from
            # QQ). Hence, we force coercion into a common parent before
            # computing the modulus. See trac #14249
            D = (polys[0],-polys[1]-(h+polys[0]) % (polys[0]))
        return JacobianMorphism_divisor_class_field(X, D, check=False)

    def _add_(self, other):
        r"""
        Return a Mumford representative of the divisor ``self + other``.

        EXAMPLES::

            sage: x = GF(37)['x'].gen()
            sage: H = HyperellipticCurve(x^5 + 12*x^4 + 13*x^3 + 15*x^2 + 33*x)
            sage: J = H.jacobian()(GF(37))

        ::

            sage: P1 = J(H.lift_x(2)); P1
            (x + 35, y + 26)
            sage: P1 + P1  # indirect doctest
            (x^2 + 33*x + 4, y + 13*x)
        """
        X = self.parent()
        C = X.curve()
        f, h = C.hyperelliptic_polynomials()
        genus = C.genus()
        if h == 0:
            D = cantor_composition_simple(self.__polys, other.__polys, f, genus)
            if D[0].degree() > genus:
                D = cantor_reduction_simple(D[0], D[1], f, genus)
        else:
            D = cantor_composition(self.__polys, other.__polys, f, h, genus)
            if D[0].degree() > genus:
                D = cantor_reduction(D[0], D[1], f, h, genus)
        return JacobianMorphism_divisor_class_field(X, D, check=False)

    def _sub_(self, other):
        r"""
        Return a Mumford representative of the divisor ``self - other``.

        EXAMPLES::

            sage: x = GF(37)['x'].gen()
            sage: H = HyperellipticCurve(x^5 + 12*x^4 + 13*x^3 + 15*x^2 + 33*x)
            sage: J = H.jacobian()(GF(37))

        ::

            sage: P1 = J(H.lift_x(2)); P1
            (x + 35, y + 26)
            sage: P1 - P1  # indirect doctest
            (1)

        ::

            sage: P2 = J(H.lift_x(4)); P2
            (x + 33, y + 34)

        Observe that the `x`-coordinates are the same but the
        `y`-coordinates differ::

            sage: P1 - P2  # indirect doctest
            (x^2 + 31*x + 8, y + 7*x + 12)
            sage: P1 + P2  # indirect doctest
            (x^2 + 31*x + 8, y + 4*x + 18)
            sage: (P1 - P2) - (P1 + P2) + 2*P2  # indirect doctest
            (1)
        """
        return self + (-other)



# sage_setup: distribution = sagemath-flint
# distutils: libraries = flint
# distutils: depends = flint/ca_field.h

################################################################################
# This file is auto-generated by the script
#   SAGE_ROOT/src/sage_setup/autogen/flint_autogen.py.
# From the commit 3e2c3a3e091106a25ca9c6fba28e02f2cbcd654a
# Do not modify by hand! Fix and rerun the script instead.
################################################################################

from libc.stdio cimport FILE
from sage.libs.gmp.types cimport *
from sage.libs.mpfr.types cimport *
from sage.libs.flint.types cimport *

cdef extern from "flint_wrap.h":
    void ca_field_init_qq(ca_field_t K, ca_ctx_t ctx) noexcept
    void ca_field_init_nf(ca_field_t K, const qqbar_t x, ca_ctx_t ctx) noexcept
    void ca_field_init_const(ca_field_t K, calcium_func_code func, ca_ctx_t ctx) noexcept
    void ca_field_init_fx(ca_field_t K, calcium_func_code func, const ca_t x, ca_ctx_t ctx) noexcept
    void ca_field_init_fxy(ca_field_t K, calcium_func_code func, const ca_t x, const ca_t y, ca_ctx_t ctx) noexcept
    void ca_field_init_multi(ca_field_t K, slong len, ca_ctx_t ctx) noexcept
    void ca_field_set_ext(ca_field_t K, slong i, ca_ext_srcptr x_index, ca_ctx_t ctx) noexcept
    void ca_field_clear(ca_field_t K, ca_ctx_t ctx) noexcept
    void ca_field_print(const ca_field_t K, ca_ctx_t ctx) noexcept
    void ca_field_build_ideal(ca_field_t K, ca_ctx_t ctx) noexcept
    void ca_field_build_ideal_erf(ca_field_t K, ca_ctx_t ctx) noexcept
    int ca_field_cmp(const ca_field_t K1, const ca_field_t K2, ca_ctx_t ctx) noexcept
    void ca_field_cache_init(ca_field_cache_t cache, ca_ctx_t ctx) noexcept
    void ca_field_cache_clear(ca_field_cache_t cache, ca_ctx_t ctx) noexcept
    ca_field_ptr ca_field_cache_insert_ext(ca_field_cache_t cache, ca_ext_struct ** x, slong len, ca_ctx_t ctx) noexcept



# sage_setup: distribution = sagemath-modules
r"""
Hermite form computation for function fields

This module provides an optimized implementation of the algorithm computing
Hermite forms of matrices over polynomials. This is the workhorse of the
function field machinery of Sage.

EXAMPLES::

    sage: P.<x> = PolynomialRing(QQ)
    sage: A = matrix(P,3,[-(x-1)^((i-j+1) % 3) for i in range(3) for j in range(3)])
    sage: A
    [        -x + 1             -1 -x^2 + 2*x - 1]
    [-x^2 + 2*x - 1         -x + 1             -1]
    [            -1 -x^2 + 2*x - 1         -x + 1]
    sage: from sage.rings.function_field.hermite_form_polynomial import reversed_hermite_form
    sage: B = copy(A)
    sage: U = reversed_hermite_form(B, transformation=True)
    sage: U * A == B
    True
    sage: B
    [x^3 - 3*x^2 + 3*x - 2                     0                     0]
    [                    0 x^3 - 3*x^2 + 3*x - 2                     0]
    [        x^2 - 2*x + 1                 x - 1                     1]

The function :func:`reversed_hermite_form` computes the reversed hermite form,
which is reversed both row-wise and column-wise from the usual hermite form.
Let us check it::

    sage: A.reverse_rows_and_columns()
    sage: C = copy(A.hermite_form())
    sage: C.reverse_rows_and_columns()
    sage: C
    [x^3 - 3*x^2 + 3*x - 2                    0                     0]
    [                    0 x^3 - 3*x^2 + 3*x - 2                     0]
    [        x^2 - 2*x + 1                 x - 1                     1]
    sage: C == B
    True

AUTHORS:

- Kwankyu Lee (2021-05-21): initial version
"""

# ****************************************************************************
#       Copyright (C) 2021 Kwankyu Lee <ekwankyu@gmail.com>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#  as published by the Free Software Foundation; either version 2 of
#  the License, or (at your option) any later version.
#                  https://www.gnu.org/licenses/
# ****************************************************************************

from sage.matrix.matrix cimport Matrix
from sage.rings.polynomial.polynomial_element cimport Polynomial
from sage.matrix.constructor import identity_matrix


def reversed_hermite_form(Matrix mat, bint transformation=False):
    """
    Transform the matrix in place to reversed hermite normal form and
    optionally return the transformation matrix.

    INPUT:

    - ``transformation`` -- boolean (default: ``False``); if ``True``,
      return the transformation matrix

    EXAMPLES::

        sage: from sage.rings.function_field.hermite_form_polynomial import reversed_hermite_form
        sage: P.<x> = PolynomialRing(QQ)
        sage: A = matrix(P,3,[-(x-1)^((i-2*j) % 4) for i in range(3) for j in range(3)])
        sage: A
        [                    -1         -x^2 + 2*x - 1                     -1]
        [                -x + 1 -x^3 + 3*x^2 - 3*x + 1                 -x + 1]
        [        -x^2 + 2*x - 1                     -1         -x^2 + 2*x - 1]
        sage: B = copy(A)
        sage: U = reversed_hermite_form(B, transformation=True)
        sage: U * A == B
        True
        sage: B
        [                        0                         0                         0]
        [                        0 x^4 - 4*x^3 + 6*x^2 - 4*x                         0]
        [                        1             x^2 - 2*x + 1                         1]
    """
    cdef Matrix A = mat
    cdef Matrix U

    cdef Py_ssize_t m = A.nrows()
    cdef Py_ssize_t n = A.ncols()
    cdef Py_ssize_t i = m - 1
    cdef Py_ssize_t j = n - 1
    cdef Py_ssize_t k, l, c, ip

    cdef Polynomial a, b, d, p, q, e, f, Aic, Alc, Uic, Ulc
    cdef Polynomial zero = A.base_ring().zero()

    cdef int di, dk

    cdef list pivot_cols = []

    if transformation:
        U = <Matrix> identity_matrix(A.base_ring(), m)

    while j >= 0:
        # find the first row with nonzero entry in jth column
        k = i
        while k >= 0 and not A.get_unsafe(k, j):
            k -= 1
        if k >= 0:
            # swap the kth row with the ith row
            if k < i:
                A.swap_rows_c(i, k)
                if transformation:
                    U.swap_rows_c(i, k)
            k -= 1
            # put the row with the smalllest degree to the ith row
            di = A.get_unsafe(i, j).degree()
            while k >= 0 and A.get_unsafe(k, j):
                dk = A.get_unsafe(k, j).degree()
                if dk < di:
                    A.swap_rows_c(i, k)
                    di = dk
                    if transformation:
                        U.swap_rows_c(i, k)
                k -= 1
            l = i - 1
            while True:
                # find a row with nonzero entry in the jth column
                while l >= 0 and not A.get_unsafe(l, j):
                    l -= 1
                if l < 0:
                    break

                a = <Polynomial> A.get_unsafe(i, j)
                b = <Polynomial> A.get_unsafe(l, j)
                d, p, q = a.xgcd(b) # p * a + q * b = d = gcd(a,b)
                e = a // d
                f = -(b // d)

                A.set_unsafe(i, j, d)
                A.set_unsafe(l, j, zero)

                for c in range(j):
                    Aic = <Polynomial> A.get_unsafe(i, c)
                    Alc = <Polynomial> A.get_unsafe(l, c)
                    A.set_unsafe(i, c, p * Aic + q * Alc)
                    A.set_unsafe(l, c, f * Aic + e * Alc)
                if transformation:
                    for c in range(m):
                        Uic = <Polynomial> U.get_unsafe(i, c)
                        Ulc = <Polynomial> U.get_unsafe(l, c)
                        U.set_unsafe(i, c, p * Uic + q * Ulc)
                        U.set_unsafe(l, c, f * Uic + e * Ulc)
            pivot_cols.append(j)
            i -= 1
        j -= 1

    # reduce entries below pivots
    for i in range(len(pivot_cols)):
        j = pivot_cols[i]
        ip = m - 1 - i
        pivot = <Polynomial> A.get_unsafe(ip, j)

        # normalize the leading coefficient to one
        coeff = ~pivot.lc()
        for c in range(j + 1):
            A.set_unsafe(ip, c, <Polynomial> A.get_unsafe(ip, c) * coeff)
        if transformation:
            for c in range(m):
                U.set_unsafe(ip, c, <Polynomial> U.get_unsafe(ip, c) * coeff)

        pivot = <Polynomial> A.get_unsafe(ip,j)
        for k in range(ip + 1, m):
            q = -(<Polynomial> A.get_unsafe(k, j) // pivot)
            if q:
                for c in range(j + 1):
                    A.set_unsafe(k, c, <Polynomial> A.get_unsafe(k, c)
                                 + q * <Polynomial> A.get_unsafe(ip, c))
                if transformation:
                    for c in range(m):
                        U.set_unsafe(k, c, <Polynomial> U.get_unsafe(k, c)
                                     + q * <Polynomial> U.get_unsafe(ip, c))

    if transformation:
        return U



# sage_setup: distribution = sagemath-flint
# distutils: libraries = flint
# distutils: depends = flint/padic_mat.h

################################################################################
# This file is auto-generated by the script
#   SAGE_ROOT/src/sage_setup/autogen/flint_autogen.py.
# From the commit 3e2c3a3e091106a25ca9c6fba28e02f2cbcd654a
# Do not modify by hand! Fix and rerun the script instead.
################################################################################

from libc.stdio cimport FILE
from sage.libs.gmp.types cimport *
from sage.libs.mpfr.types cimport *
from sage.libs.flint.types cimport *

cdef extern from "flint_wrap.h":
    fmpz_mat_struct * padic_mat(const padic_mat_t A) noexcept
    fmpz * padic_mat_entry(const padic_mat_t A, slong i, slong j) noexcept
    slong padic_mat_val(const padic_mat_t A) noexcept
    slong padic_mat_prec(const padic_mat_t A) noexcept
    slong padic_mat_get_val(const padic_mat_t A) noexcept
    slong padic_mat_get_prec(const padic_mat_t A) noexcept
    slong padic_mat_nrows(const padic_mat_t A) noexcept
    slong padic_mat_ncols(const padic_mat_t A) noexcept
    void padic_mat_init(padic_mat_t A, slong r, slong c) noexcept
    void padic_mat_init2(padic_mat_t A, slong r, slong c, slong prec) noexcept
    void padic_mat_clear(padic_mat_t A) noexcept
    void _padic_mat_canonicalise(padic_mat_t A, const padic_ctx_t ctx) noexcept
    void _padic_mat_reduce(padic_mat_t A, const padic_ctx_t ctx) noexcept
    void padic_mat_reduce(padic_mat_t A, const padic_ctx_t ctx) noexcept
    bint padic_mat_is_empty(const padic_mat_t A) noexcept
    bint padic_mat_is_square(const padic_mat_t A) noexcept
    bint padic_mat_is_canonical(const padic_mat_t A, const padic_ctx_t p) noexcept
    void padic_mat_set(padic_mat_t B, const padic_mat_t A, const padic_ctx_t p) noexcept
    void padic_mat_swap(padic_mat_t A, padic_mat_t B) noexcept
    void padic_mat_swap_entrywise(padic_mat_t mat1, padic_mat_t mat2) noexcept
    void padic_mat_zero(padic_mat_t A) noexcept
    void padic_mat_one(padic_mat_t A) noexcept
    void padic_mat_set_fmpq_mat(padic_mat_t B, const fmpq_mat_t A, const padic_ctx_t ctx) noexcept
    void padic_mat_get_fmpq_mat(fmpq_mat_t B, const padic_mat_t A, const padic_ctx_t ctx) noexcept
    void padic_mat_get_entry_padic(padic_t rop, const padic_mat_t op, slong i, slong j, const padic_ctx_t ctx) noexcept
    void padic_mat_set_entry_padic(padic_mat_t rop, slong i, slong j, const padic_t op, const padic_ctx_t ctx) noexcept
    bint padic_mat_equal(const padic_mat_t A, const padic_mat_t B) noexcept
    bint padic_mat_is_zero(const padic_mat_t A) noexcept
    int padic_mat_fprint(FILE * file, const padic_mat_t A, const padic_ctx_t ctx) noexcept
    int padic_mat_fprint_pretty(FILE * file, const padic_mat_t A, const padic_ctx_t ctx) noexcept
    int padic_mat_print(const padic_mat_t A, const padic_ctx_t ctx) noexcept
    int padic_mat_print_pretty(const padic_mat_t A, const padic_ctx_t ctx) noexcept
    void padic_mat_randtest(padic_mat_t A, flint_rand_t state, const padic_ctx_t ctx) noexcept
    void padic_mat_transpose(padic_mat_t B, const padic_mat_t A) noexcept
    void _padic_mat_add(padic_mat_t C, const padic_mat_t A, const padic_mat_t B, const padic_ctx_t ctx) noexcept
    void padic_mat_add(padic_mat_t C, const padic_mat_t A, const padic_mat_t B, const padic_ctx_t ctx) noexcept
    void _padic_mat_sub(padic_mat_t C, const padic_mat_t A, const padic_mat_t B, const padic_ctx_t ctx) noexcept
    void padic_mat_sub(padic_mat_t C, const padic_mat_t A, const padic_mat_t B, const padic_ctx_t ctx) noexcept
    void _padic_mat_neg(padic_mat_t B, const padic_mat_t A) noexcept
    void padic_mat_neg(padic_mat_t B, const padic_mat_t A, const padic_ctx_t ctx) noexcept
    void _padic_mat_scalar_mul_padic(padic_mat_t B, const padic_mat_t A, const padic_t c, const padic_ctx_t ctx) noexcept
    void padic_mat_scalar_mul_padic(padic_mat_t B, const padic_mat_t A, const padic_t c, const padic_ctx_t ctx) noexcept
    void _padic_mat_scalar_mul_fmpz(padic_mat_t B, const padic_mat_t A, const fmpz_t c, const padic_ctx_t ctx) noexcept
    void padic_mat_scalar_mul_fmpz(padic_mat_t B, const padic_mat_t A, const fmpz_t c, const padic_ctx_t ctx) noexcept
    void padic_mat_scalar_div_fmpz(padic_mat_t B, const padic_mat_t A, const fmpz_t c, const padic_ctx_t ctx) noexcept
    void _padic_mat_mul(padic_mat_t C, const padic_mat_t A, const padic_mat_t B, const padic_ctx_t ctx) noexcept
    void padic_mat_mul(padic_mat_t C, const padic_mat_t A, const padic_mat_t B, const padic_ctx_t ctx) noexcept



# sage_setup: distribution = sagemath-modules
# sage.doctest: needs sage.modules sage.rings.finite_rings
r"""
Codes

Class supporting methods available for any type of code (linear, non-linear) and
over any metric (Hamming, rank).

There are further abstract classes representing certain types of codes. For
linear codes,
:class:`~sage.coding.linear_code_no_metric.AbstractLinearCodeNoMetric` contains
all the methods that any linear code can use regardless of its metric.
Inheriting from this class are base classes for linear codes over specific
metrics. For example, :class:`~sage.coding.linear_code.AbstractLinearCode` is a
base class for all linear codes over the Hamming metric.

Take the class :class:`~sage.coding.hamming_code.HammingCode`. This
class inherits from :class:`~sage.coding.linear_code.AbstractLinearCode`, since
it is a linear code over the Hamming metric.
:class:`~sage.coding.linear_code.AbstractLinearCode` then inherits from
:class:`~sage.coding.linear_code_no_metric.AbstractLinearCodeNoMetric`, since it
is a linear code. Finally, this class inherits from
:class:`~sage.coding.abstract_code.AbstractCode`, since it is a code.


The following diagram shows the inheritance relationship in the coding module::

    AbstractCode
    + AbstractLinearCodeNoMetric
    | + AbstractLinearCode
    | | + ParityCheckCode
    | | + HammingCode
    | | + CyclicCode
    | | + BCHCode
    | | + GolayCode
    | | + ReedMullerCode
    | | + GeneralizedReedSolomonCode
    | | + GoppaCode
    | + AbstractLinearRankMetricCode

Any class inheriting from AbstractCode can use the encode/decode framework.

The encoder/decoder framework within the coding module offers the creation and
use of encoders/decoders independently of codes. An encoder encodes a message
into a codeword. A decoder decodes a word into a codeword or a message,
possibly with error-correction.

Instead of creating specific encoders/decoders for every code family, some
encoders/decoders can be used by multiple code families. The encoder/decoder
framework enables just that. For example,
:class:`~sage.coding.linear_code.LinearCodeGeneratorMatrixEncoder`
can be used by any code that has a generator matrix.  Similarly,
:class:`~sage.coding.linear_code.LinearCodeNearestNeighborDecoder` can be used
for any linear code with Hamming metric.

When creating a new code family, investigate the encoder/decoder catalogs,
``codes.encoders`` and ``codes.decoders``, to see if there are suitable
encoders/decoders for your code family already implemented. If this is the case,
follow the instructions in :class:`AbstractCode` to set these up.

A new encoder must have the following methods:

- ``encode`` -- method encoding a message into a codeword
- ``unencode`` -- method decoding a codeword into a message
- ``message_space`` -- ambient space of messages that can be encoded
- ``code`` -- code of the encoder

For more information about the Encoder class, see
:class:`~sage.coding.encoder.Encoder`

A new decoder must have the following methods:

- ``decode_to_code`` or ``decode_to_message`` -- method decoding a word from the
  input space into either a codeword or a message
- ``input_space`` -- ambient space of words that can be decoded
- ``code`` -- code of the decoder

For more information about the Decoder class, see
:class:`~sage.coding.decoder.Decoder`
"""

from sage.structure.parent import Parent
from sage.misc.cachefunc import cached_method
from copy import copy
from sage.rings.integer import Integer

import inspect
from sage.misc.sageinspect import sage_getargspec


def _explain_constructor(cl):
    r"""
    Internal function for use error messages when constructing encoders and decoders.

    EXAMPLES::

        sage: from sage.coding.linear_code import LinearCodeSyndromeDecoder
        sage: from sage.coding.abstract_code import _explain_constructor
        sage: cl = LinearCodeSyndromeDecoder
        sage: _explain_constructor(cl)
        "The constructor requires no arguments.\nIt takes the optional
        arguments ['maximum_error_weight'].\nSee the documentation of
        sage.coding.linear_code.LinearCodeSyndromeDecoder for more details."

        sage: from sage.coding.information_set_decoder import LinearCodeInformationSetDecoder
        sage: cl = LinearCodeInformationSetDecoder
        sage: _explain_constructor(cl)
        "The constructor requires the arguments ['number_errors'].\nIt takes the optional arguments ['algorithm'].\nIt accepts unspecified arguments as well.\nSee the documentation of sage.coding.information_set_decoder.LinearCodeInformationSetDecoder for more details."
    """
    if inspect.isclass(cl):
        argspec = sage_getargspec(cl.__init__)
        skip = 2 # skip the self and code arguments
    else:
        # Not a class, assume it's a factory function posing as a class
        argspec = sage_getargspec(cl)
        skip = 1 # skip code argument
    if argspec.defaults:
        args = argspec.args[skip:-len(argspec.defaults)]
        kwargs = argspec.args[-len(argspec.defaults):]
        opts = "It takes the optional arguments {}.".format(kwargs)
    else:
        args = argspec.args[skip:]
        opts = "It takes no optional arguments."
    if args:
        reqs = "The constructor requires the arguments {}.".format(args)
    else:
        reqs = "The constructor requires no arguments."
    if argspec.varargs or argspec.varkw:
        var = "It accepts unspecified arguments as well.\n"
    else:
        var = ""
    return ("{}\n{}\n{}See the documentation of {}.{} for more details."
            .format(reqs, opts, var, cl.__module__, cl.__name__))


class AbstractCode(Parent):
    r"""
    Abstract class for codes.

    This class contains all the methods that can be used on any code
    and on any code family. As opposed to
    :class:`sage.coding.linear_code.AbstractLinearCode`, this class makes no
    assumptions about linearity, metric, finiteness or the number of alphabets.

    The abstract notion of "code" that is implicitly used for this class is any
    enumerable subset of a cartesian product `A_1 \times A_2 \times \ldots
    \times A_n` for some sets `A_i`. Note that this class makes no attempt to
    directly represent the code in this fashion, allowing subclasses to make the
    appropriate choices. The notion of metric is also not mathematically
    enforced in any way, and is simply stored as a string value.

    Every code-related class should inherit from this abstract class.

    To implement a code, you need to:

    - inherit from :class:`AbstractCode`

    - call :class:`AbstractCode` ``__init__`` method in the subclass constructor.
      Example: ``super().__init__(length, "EncoderName",
      "DecoderName", "metric")``. "EncoderName" and "DecoderName" are set to
      ``None`` by default, a generic code class such as AbstractCode does
      not necessarily have to have general encoders/decoders. However, if you
      want to use the encoding/decoding methods, you have to add these.

    - since this class does not specify any category, it is highly recommended
      to set up the category framework in the subclass. To do this, use the
      ``Parent.__init__(self, base, facade, category)`` function in the subclass
      constructor. A good example is in
      :class:`sage.coding.linear_code.AbstractLinearCode`.

    - it is also recommended to override the ``ambient_space`` method, which is
      required by ``__call__``

    - to use the encoder/decoder framework, one has to set up the category and
      related functions ``__iter__`` and ``__contains__``. A good example is in
      :class:`sage.coding.linear_code.AbstractLinearCode`.

    - add the following two lines on the class level::

          _registered_encoders = {}
          _registered_decoders = {}


    - fill the dictionary of its encoders in ``sage.coding.__init__.py`` file.
      Example: I want to link the encoder ``MyEncoderClass`` to ``MyNewCodeClass``
      under the name ``MyEncoderName``.
      All I need to do is to write this line in the ``__init__.py`` file:
      ``MyNewCodeClass._registered_encoders["NameOfMyEncoder"] = MyEncoderClass``
      and all instances of ``MyNewCodeClass`` will be able to use instances of
      ``MyEncoderClass``.

    - fill the dictionary of its decoders in ``sage.coding.__init__`` file.
      Example: I want to link the encoder ``MyDecoderClass`` to ``MyNewCodeClass``
      under the name ``MyDecoderName``.
      All I need to do is to write this line in the ``__init__.py`` file:
      ``MyNewCodeClass._registered_decoders["NameOfMyDecoder"] = MyDecoderClass``
      and all instances of ``MyNewCodeClass`` will be able to use instances of
      ``MyDecoderClass``.


    As the class :class:`AbstractCode` is not designed to be instantiated, it
    does not have any representation methods. You should implement ``_repr_``
    and ``_latex_`` methods in the subclass.
    """

    def __init__(self, length, default_encoder_name=None,
                 default_decoder_name=None, metric='Hamming') -> None:
        r"""
        Initialize mandatory parameters that any code shares.

        This method only exists for inheritance purposes as it initializes
        parameters that need to be known by every code. The class
        :class:`sage.coding.abstract_code.AbstractCode` should never be
        directly instantiated.

        INPUT:

        - ``length`` -- the length of ``self`` (a Python int or a Sage Integer,
          must be >= 0)

        - ``default_encoder_name`` -- (default: ``None``) the name of
          the default encoder of ``self``

        - ``default_decoder_name`` -- (default: ``None``) the name of
          the default decoder of ``self``

        - ``metric`` -- (default: ``Hamming``) the name of the metric of ``self``

        EXAMPLES:

        The following example demonstrates how to use a subclass of ``AbstractCode``
        for representing a new family of codes::

            sage: from sage.coding.abstract_code import AbstractCode
            sage: class MyCodeFamily(AbstractCode):
            ....:   def __init__(self, length):
            ....:       super().__init__(length)
            ....:   def __iter__(self):
            ....:       for i in range(self.length() + 1):
            ....:            yield vector([1 for j in range(i)] + [0 for k in range(i, self.length())])
            ....:   def __contains__(self, word):
            ....:       return word in list(self)
            ....:   def _repr_(self):
            ....:       return "Dummy code of length {}".format(self.length())

        We now instantiate a member of our newly made code family::

            sage: C = MyCodeFamily(6)

        We can check its existence and parameters::

            sage: C
            Dummy code of length 6

        We can list its elements and check if an element is in the code::

            sage: list(C)
            [(0, 0, 0, 0, 0, 0),
            (1, 0, 0, 0, 0, 0),
            (1, 1, 0, 0, 0, 0),
            (1, 1, 1, 0, 0, 0),
            (1, 1, 1, 1, 0, 0),
            (1, 1, 1, 1, 1, 0),
            (1, 1, 1, 1, 1, 1)]
            sage: vector((0, 1, 0, 0, 0, 1)) in C
            False
            sage: vector((1, 1, 1, 0, 0, 0)) in C
            True

        And coming from AbstractCode code::

            sage: C.metric()
            'Hamming'

        TESTS:

        If the length field is neither a Python int nor a Sage Integer, it will
        raise a exception::

            sage: C = MyCodeFamily(10.0)
            Traceback (most recent call last):
            ...
            ValueError: length must be a Python int or a Sage Integer

        If the length of the code is negative, it will raise an exception::

            sage: C = MyCodeFamily(-1)
            Traceback (most recent call last):
            ...
            ValueError: length must be a non-negative integer

        Codes of length 0 are allowed::

            sage: C = MyCodeFamily(0)
            sage: C.length()
            0
        """

        if not isinstance(length, (int, Integer)):
            raise ValueError("length must be a Python int or a Sage Integer")
        if length < 0:
            raise ValueError("length must be a non-negative integer")

        self._length = length
        self._metric = metric

        self._default_decoder_name = default_decoder_name
        self._default_encoder_name = default_encoder_name
        if not self._default_decoder_name:
            self._registered_encoders = {}
        if not self._default_encoder_name:
            self._registered_decoders = {}

    def __getstate__(self):
        """
        Used for pickling codes.

        TESTS::

            sage: C = codes.HammingCode(GF(2), 3)
            sage: '_registered_encoders' in C.__getstate__()
            True
        """
        d = super().__getstate__()
        d['_registered_encoders'] = self._registered_encoders
        d['_registered_decoders'] = self._registered_decoders
        return d

    def __iter__(self):
        r"""
        Return an error message requiring to override ``__iter__`` in ``self``.

        As one has to implement specific category related methods (``__iter__`` and
        ``__contains__``) when writing a new code class which inherits from
        :class:`AbstractCode`, the generic call to ``__iter__`` has to fail.

        EXAMPLES:

        We create a new code class::

            sage: from sage.coding.abstract_code import AbstractCode
            sage: class MyCode(AbstractCode):
            ....:    def __init__(self):
            ....:        super().__init__(10)

        We check we get a sensible error message while asking for an
        iterator over the elements of our new class::

            sage: C = MyCode()
            sage: list(C)
            Traceback (most recent call last):
            ...
            RuntimeError: Please override __iter__ in the implementation of <class '__main__.MyCode'>
        """
        raise RuntimeError("Please override __iter__ in the implementation of {}".format(self.parent()))

    def __contains__(self, c) -> bool:
        r"""
        Return an error message requiring to override ``__contains__`` in ``self``.

        As one has to implement specific category related methods (``__iter__`` and
        ``__contains__``) when writing a new code class which inherits from
        :class:`AbstractCode`, the generic call to ``__contains__`` has to fail.

        EXAMPLES:

        We create a new code class::

            sage: from sage.coding.abstract_code import AbstractCode
            sage: class MyCode(AbstractCode):
            ....:    def __init__(self, length):
            ....:        super().__init__(length)

        We check we get a sensible error message while asking if an element is
        in our new class::

            sage: C = MyCode(3)
            sage: vector((1, 0, 0, 0, 0, 1, 1)) in C
            Traceback (most recent call last):
            ...
            RuntimeError: Please override __contains__ in the implementation of <class '__main__.MyCode'>
        """
        raise RuntimeError("Please override __contains__ in the implementation of {}".format(self.parent()))

    def ambient_space(self):
        r"""
        Return an error stating ``ambient_space`` of ``self`` is not implemented.

        This method is required by :meth:`__call__`.

        EXAMPLES::

            sage: from sage.coding.abstract_code import AbstractCode
            sage: class MyCode(AbstractCode):
            ....:    def __init__(self, length):
            ....:        super().__init__(length)
            sage: C = MyCode(3)
            sage: C.ambient_space()
            Traceback (most recent call last):
            ...
            NotImplementedError: No ambient space implemented for this code.
        """
        raise NotImplementedError("No ambient space implemented for this code.")

    def __call__(self, m):
        r"""
        Return either ``m`` if it is a codeword or ``self.encode(m)``
        if it is an element of the message space of the encoder used by
        ``encode``.

        This implementation depends on :meth:`ambient_space`.

        INPUT:

        - ``m`` -- a vector whose length equals to code's length or an element
          of the message space used by ``encode``

        - ``**kwargs`` -- extra arguments are forwarded to ``encode``

        EXAMPLES::

            sage: G = Matrix(GF(2), [[1,1,1,0,0,0,0],[1,0,0,1,1,0,0],[0,1,0,1,0,1,0],[1,1,0,1,0,0,1]])
            sage: C = LinearCode(G)
            sage: word = vector((0, 1, 1, 0))
            sage: C(word)
            (0, 1, 1, 0, 0, 1, 1)

            sage: c = C.random_element()
            sage: C(c) == c
            True

        TESTS:

        If one passes a vector which belongs to the ambient space, it has to be a codeword.
        Otherwise, an exception is raised::

            sage: G = Matrix(GF(2), [[1,1,1,0,0,0,0],[1,0,0,1,1,0,0],[0,1,0,1,0,1,0],[1,1,0,1,0,0,1]])
            sage: C = LinearCode(G)
            sage: word = vector((0, 1, 1, 0, 0, 1, 0))
            sage: C(word)
            Traceback (most recent call last):
            ...
            ValueError: If the input is a vector which belongs to the ambient space, it has to be a codeword
        """
        if m in self.ambient_space():
            if m in self:
                return m
            else:
                raise ValueError("If the input is a vector which belongs to the ambient space, it has to be a codeword")
        else:
            return self.encode(m)

    def _repr_(self):
        r"""
        Return an error message requiring to override ``_repr_`` in ``self``.

        As one has to implement specific representation methods (``_repr_`` and
        ``_latex_``) when writing a new code class which inherits from
        :class:`AbstractCode`, the generic call to ``_repr_`` has to fail.

        EXAMPLES:

        We create a new code class::

            sage: from sage.coding.abstract_code import AbstractCode
            sage: class MyCode(AbstractCode):
            ....:    def __init__(self):
            ....:        super().__init__(10)

        We check we get a sensible error message while asking for a string
        representation of an instance of our new class::

            sage: C = MyCode()
            sage: C #random
            Traceback (most recent call last):
            ...
            RuntimeError: Please override _repr_ in the implementation of <class '__main__.MyCode'>
        """
        raise RuntimeError("Please override _repr_ in the implementation of {}".format(self.parent()))

    def _latex_(self):
        r"""
        Return an error message requiring to override ``_latex_`` in ``self``.

        As one has to implement specific representation methods (``_repr_`` and
        ``_latex_``) when writing a new code class which inherits from
        :class:`AbstractCode`, the generic call to ``_latex_`` has to fail.

        EXAMPLES:

        We create a new code class::

            sage: from sage.coding.abstract_code import AbstractCode
            sage: class MyCode(AbstractCode):
            ....:    def __init__(self):
            ....:        super().__init__(10)

        We check we get a sensible error message while asking for a string
        representation of an instance of our new class::

            sage: C = MyCode()
            sage: latex(C)
            Traceback (most recent call last):
            ...
            RuntimeError: Please override _latex_ in the implementation of <class '__main__.MyCode'>
        """
        raise RuntimeError("Please override _latex_ in the implementation of {}".format(self.parent()))

    def list(self):
        r"""
        Return a list of all elements of this code.

        EXAMPLES::

            sage: C = codes.HammingCode(GF(2), 3)
            sage: Clist = C.list()
            sage: Clist[5]; Clist[5] in C
            (1, 0, 1, 0, 1, 0, 1)
            True
        """
        return list(self)

    def length(self):
        r"""
        Return the length of this code.

        EXAMPLES::

            sage: C = codes.HammingCode(GF(2), 3)
            sage: C.length()
            7
        """
        return self._length

    def metric(self):
        """
        Return the metric of ``self``.

        EXAMPLES::

            sage: C = codes.HammingCode(GF(2), 3)
            sage: C.metric()
            'Hamming'
        """
        return self._metric

###################### Encoding-Decoding #######################################

    def add_decoder(self, name, decoder):
        r"""
        Add an decoder to the list of registered decoders of ``self``.

        .. NOTE::

            This method only adds ``decoder`` to ``self``, and not to any member of the class
            of ``self``. To know how to add an :class:`sage.coding.decoder.Decoder`, please refer
            to the documentation of :class:`AbstractCode`.

        INPUT:

        - ``name`` -- the string name for the decoder

        - ``decoder`` -- the class name of the decoder

        EXAMPLES:

        First of all, we create a (very basic) new decoder::

            sage: class MyDecoder(sage.coding.decoder.Decoder):
            ....:   def __init__(self, code):
            ....:       super().__init__(code)
            ....:   def _repr_(self):
            ....:       return "MyDecoder decoder with associated code %s" % self.code()

        We now create a new code::

            sage: C = codes.HammingCode(GF(2), 3)

        We can add our new decoder to the list of available decoders of C::

            sage: C.add_decoder("MyDecoder", MyDecoder)
            sage: sorted(C.decoders_available())
            ['InformationSet', 'MyDecoder', 'NearestNeighbor', 'Syndrome']

        We can verify that any new code will not know MyDecoder::

            sage: C2 = codes.HammingCode(GF(2), 3)
            sage: sorted(C2.decoders_available())
            ['InformationSet', 'NearestNeighbor', 'Syndrome']

        TESTS:

        It is impossible to use a name which is in the dictionary of available decoders::

            sage: C.add_decoder("Syndrome", MyDecoder)
            Traceback (most recent call last):
            ...
            ValueError: There is already a registered decoder with this name
        """
        if self._registered_decoders == self.__class__._registered_decoders:
            self._registered_decoders = copy(self._registered_decoders)
            reg_dec = self._registered_decoders
            if name in reg_dec:
                raise ValueError("There is already a registered decoder with this name")
            reg_dec[name] = decoder
        else:
            if name in self._registered_decoders:
                raise ValueError("There is already a registered decoder with this name")
            reg_dec[name] = decoder

    def add_encoder(self, name, encoder):
        r"""
        Add an encoder to the list of registered encoders of ``self``.

        .. NOTE::

            This method only adds ``encoder`` to ``self``, and not to any member of the class
            of ``self``. To know how to add an :class:`sage.coding.encoder.Encoder`, please refer
            to the documentation of :class:`AbstractCode`.

        INPUT:

        - ``name`` -- the string name for the encoder

        - ``encoder`` -- the class name of the encoder

        EXAMPLES:

        First of all, we create a (very basic) new encoder::

            sage: class MyEncoder(sage.coding.encoder.Encoder):
            ....:   def __init__(self, code):
            ....:       super().__init__(code)
            ....:   def _repr_(self):
            ....:       return "MyEncoder encoder with associated code %s" % self.code()

        We now create a new code::

            sage: C = codes.HammingCode(GF(2), 3)

        We can add our new encoder to the list of available encoders of C::

            sage: C.add_encoder("MyEncoder", MyEncoder)
            sage: sorted(C.encoders_available())
            ['MyEncoder', 'Systematic']

        We can verify that any new code will not know MyEncoder::

            sage: C2 = codes.HammingCode(GF(2), 3)
            sage: sorted(C2.encoders_available())
            ['Systematic']

        TESTS:

        It is impossible to use a name which is in the dictionary of available encoders::

            sage: C.add_encoder("Systematic", MyEncoder)
            Traceback (most recent call last):
            ...
            ValueError: There is already a registered encoder with this name
        """
        if self._registered_encoders == self.__class__._registered_encoders:
            self._registered_encoders = copy(self._registered_encoders)
            reg_enc = self._registered_encoders
            if name in reg_enc:
                raise ValueError("There is already a registered encoder with this name")
            reg_enc[name] = encoder
        else:
            if name in self._registered_encoders:
                raise ValueError("There is already a registered encoder with this name")
            reg_enc[name] = encoder

    def decode_to_code(self, word, decoder_name=None, *args, **kwargs):
        r"""
        Correct the errors in ``word`` and returns a codeword.

        INPUT:

        - ``word`` -- an element in the ambient space as ``self``

        - ``decoder_name`` -- (default: ``None``) name of the decoder which will be used
          to decode ``word``. The default decoder of ``self`` will be used if
          default value is kept.

        - ``args``, ``kwargs`` -- all additional arguments are forwarded to :meth:`decoder`

        OUTPUT: a vector of ``self``

        EXAMPLES::

            sage: G = Matrix(GF(2), [[1,1,1,0,0,0,0], [1,0,0,1,1,0,0],
            ....:                    [0,1,0,1,0,1,0], [1,1,0,1,0,0,1]])
            sage: C = LinearCode(G)
            sage: word = vector(GF(2), (1, 1, 0, 0, 1, 1, 0))
            sage: w_err = word + vector(GF(2), (1, 0, 0, 0, 0, 0, 0))
            sage: C.decode_to_code(w_err)
            (1, 1, 0, 0, 1, 1, 0)

        It is possible to manually choose the decoder amongst the list of the available ones::

            sage: sorted(C.decoders_available())
            ['InformationSet', 'NearestNeighbor', 'Syndrome']
            sage: C.decode_to_code(w_err, 'NearestNeighbor')
            (1, 1, 0, 0, 1, 1, 0)
        """
        D = self.decoder(decoder_name, *args, **kwargs)
        return D.decode_to_code(word)

    def decode_to_message(self, word, decoder_name=None, *args, **kwargs):
        r"""
        Correct the errors in word and decodes it to the message space.

        INPUT:

        - ``word`` -- an element in the ambient space as ``self``

        - ``decoder_name`` -- (default: ``None``) name of the decoder which will be used
          to decode ``word``. The default decoder of ``self`` will be used if
          default value is kept.

        - ``args``, ``kwargs`` -- all additional arguments are forwarded to :meth:`decoder`

        OUTPUT: a vector of the message space of ``self``

        EXAMPLES::

            sage: G = Matrix(GF(2), [[1,1,1,0,0,0,0], [1,0,0,1,1,0,0],
            ....:                    [0,1,0,1,0,1,0], [1,1,0,1,0,0,1]])
            sage: C = LinearCode(G)
            sage: word = vector(GF(2), (1, 1, 0, 0, 1, 1, 0))
            sage: C.decode_to_message(word)
            (1, 1, 0, 0)

        It is possible to manually choose the decoder amongst the list of the available ones::

            sage: sorted(C.decoders_available())
            ['InformationSet', 'NearestNeighbor', 'Syndrome']
            sage: C.decode_to_message(word, 'NearestNeighbor')
            (1, 1, 0, 0)
        """
        return self.unencode(self.decode_to_code(word, decoder_name, *args, **kwargs), **kwargs)

    @cached_method
    def decoder(self, decoder_name=None, *args, **kwargs):
        r"""
        Return a decoder of ``self``.

        INPUT:

        - ``decoder_name`` -- (default: ``None``) name of the decoder which will be
          returned. The default decoder of ``self`` will be used if
          default value is kept.

        - ``args``, ``kwargs`` -- all additional arguments will be forwarded to the constructor of the decoder
          that will be returned by this method

        OUTPUT: a decoder object

        Besides creating the decoder and returning it, this method also stores
        the decoder in a cache. With this behaviour, each decoder will be created
        at most one time for ``self``.

        EXAMPLES::

            sage: G = Matrix(GF(2), [[1,1,1,0,0,0,0], [1,0,0,1,1,0,0],
            ....:                    [0,1,0,1,0,1,0], [1,1,0,1,0,0,1]])
            sage: C = LinearCode(G)
            sage: C.decoder()
            Syndrome decoder for [7, 4] linear code over GF(2) handling errors of weight up to 1

        If there is no decoder for the code, we return an error::

            sage: from sage.coding.abstract_code import AbstractCode
            sage: class MyCodeFamily(AbstractCode):
            ....:   def __init__(self, length, field):
            ....:       sage.coding.abstract_code.AbstractCode.__init__(self, length)
            ....:       Parent.__init__(self, base=field, facade=False, category=Sets())
            ....:       self._field = field
            ....:   def field(self):
            ....:       return self._field
            ....:   def _repr_(self):
            ....:       return "%d dummy code over GF(%s)" % (self.length(), self.field().cardinality())
            sage: D = MyCodeFamily(5, GF(2))
            sage: D.decoder()
            Traceback (most recent call last):
            ...
            NotImplementedError: No decoder implemented for this code.

        If the name of a decoder which is not known by ``self`` is passed,
        an exception will be raised::

            sage: sorted(C.decoders_available())
            ['InformationSet', 'NearestNeighbor', 'Syndrome']
            sage: C.decoder('Try')
            Traceback (most recent call last):
            ...
            ValueError: There is no Decoder named 'Try'.
            The known Decoders are: ['InformationSet', 'NearestNeighbor', 'Syndrome']

        Some decoders take extra arguments. If the user forgets to supply these,
        the error message attempts to be helpful::

            sage: C.decoder('InformationSet')
            Traceback (most recent call last):
            ...
            ValueError: Constructing the InformationSet decoder failed,
            possibly due to missing or incorrect parameters.
            The constructor requires the arguments ['number_errors'].
            It takes the optional arguments ['algorithm'].
            It accepts unspecified arguments as well. See the documentation of
            sage.coding.information_set_decoder.LinearCodeInformationSetDecoder
            for more details.
        """
        if not self._default_decoder_name:
            raise NotImplementedError("No decoder implemented for this code.")
        if decoder_name is None:
            decoder_name = self._default_decoder_name
        if decoder_name in self._registered_decoders:
            decClass = self._registered_decoders[decoder_name]
            try:
                return decClass(self, *args, **kwargs)
            except TypeError:
                raise ValueError(
                        "Constructing the {0} decoder failed, possibly due "
                        "to missing or incorrect parameters.\n{1}".format(
                            decoder_name, _explain_constructor(decClass)))
        else:
            raise ValueError(
                    "There is no Decoder named '{0}'. The known Decoders are: "
                    "{1}".format(decoder_name, self.decoders_available()))

    def decoders_available(self, classes=False):
        r"""
        Return a list of the available decoders' names for ``self``.

        INPUT:

        - ``classes`` -- boolean (default: ``False``); if ``classes`` is set to
          ``True``, return instead a :class:`dict` mapping available decoder
          name to the associated decoder class

        OUTPUT: list of strings, or a :class:`dict` mapping strings to classes

        EXAMPLES::

            sage: G = Matrix(GF(2), [[1,1,1,0,0,0,0], [1,0,0,1,1,0,0],
            ....:                    [0,1,0,1,0,1,0], [1,1,0,1,0,0,1]])
            sage: C = LinearCode(G)
            sage: C.decoders_available()
            ['InformationSet', 'NearestNeighbor', 'Syndrome']

            sage: dictionary = C.decoders_available(True)
            sage: sorted(dictionary.keys())
            ['InformationSet', 'NearestNeighbor', 'Syndrome']
            sage: dictionary['NearestNeighbor']
            <class 'sage.coding.linear_code.LinearCodeNearestNeighborDecoder'>
        """
        if classes:
            return copy(self._registered_decoders)

        return sorted(self._registered_decoders)

    def encode(self, word, encoder_name=None, *args, **kwargs):
        r"""
        Transform an element of a message space into a codeword.

        INPUT:

        - ``word`` -- an element of a message space of the code

        - ``encoder_name`` -- (default: ``None``) name of the encoder which
          will be used to encode ``word``. The default encoder of ``self`` will
          be used if default value is kept.

        - ``args``, ``kwargs`` -- all additional arguments are forwarded to the
          construction of the encoder that is used

        One can use the following shortcut to encode a word ::

            C(word)

        OUTPUT: a vector of ``self``

        EXAMPLES::

            sage: G = Matrix(GF(2), [[1,1,1,0,0,0,0], [1,0,0,1,1,0,0],
            ....:                    [0,1,0,1,0,1,0], [1,1,0,1,0,0,1]])
            sage: C = LinearCode(G)
            sage: word = vector((0, 1, 1, 0))
            sage: C.encode(word)
            (0, 1, 1, 0, 0, 1, 1)
            sage: C(word)
            (0, 1, 1, 0, 0, 1, 1)

        It is possible to manually choose the encoder amongst the list of the available ones::

            sage: sorted(C.encoders_available())
            ['GeneratorMatrix', 'Systematic']
            sage: word = vector((0, 1, 1, 0))
            sage: C.encode(word, 'GeneratorMatrix')
            (0, 1, 1, 0, 0, 1, 1)
        """
        E = self.encoder(encoder_name, *args, **kwargs)
        return E.encode(word)

    @cached_method
    def encoder(self, encoder_name=None, *args, **kwargs):
        r"""
        Return an encoder of ``self``.

        The returned encoder provided by this method is cached.

        This methods creates a new instance of the encoder subclass designated by ``encoder_name``.
        While it is also possible to do the same by directly calling the subclass' constructor,
        it is strongly advised to use this method to take advantage of the caching mechanism.

        INPUT:

        - ``encoder_name`` -- (default: ``None``) name of the encoder which will be
          returned. The default encoder of ``self`` will be used if
          default value is kept.

        - ``args``, ``kwargs`` -- all additional arguments are forwarded to the
          constructor of the encoder this method will return

        OUTPUT: an Encoder object

        .. NOTE::

            The default encoder always has `F^{k}` as message space, with `k` the dimension
            of ``self`` and `F` the base ring of ``self``.

        EXAMPLES::

            sage: G = Matrix(GF(2), [[1,1,1,0,0,0,0], [1,0,0,1,1,0,0],
            ....:                    [0,1,0,1,0,1,0], [1,1,0,1,0,0,1]])
            sage: C = LinearCode(G)
            sage: C.encoder()
            Generator matrix-based encoder for [7, 4] linear code over GF(2)

        If there is no encoder for the code, we return an error::

            sage: from sage.coding.abstract_code import AbstractCode
            sage: class MyCodeFamily(AbstractCode):
            ....:   def __init__(self, length, field):
            ....:       sage.coding.abstract_code.AbstractCode.__init__(self, length)
            ....:       Parent.__init__(self, base=field, facade=False, category=Sets())
            ....:       self._field = field
            ....:   def field(self):
            ....:       return self._field
            ....:   def _repr_(self):
            ....:       return "%d dummy code over GF(%s)" % (self.length(),
            ....:                                             self.field().cardinality())
            sage: D = MyCodeFamily(5, GF(2))
            sage: D.encoder()
            Traceback (most recent call last):
            ...
            NotImplementedError: No encoder implemented for this code.

        We check that the returned encoder is cached::

            sage: C.encoder.is_in_cache()
            True

        If the name of an encoder which is not known by ``self`` is passed,
        an exception will be raised::

            sage: sorted(C.encoders_available())
            ['GeneratorMatrix', 'Systematic']
            sage: C.encoder('NonExistingEncoder')
            Traceback (most recent call last):
            ...
            ValueError: There is no Encoder named 'NonExistingEncoder'.
            The known Encoders are: ['GeneratorMatrix', 'Systematic']

        Some encoders take extra arguments. If the user incorrectly supplies
        these, the error message attempts to be helpful::

            sage: C.encoder('Systematic', strange_parameter=True)
            Traceback (most recent call last):
            ...
            ValueError: Constructing the Systematic encoder failed,
            possibly due to missing or incorrect parameters.
            The constructor requires no arguments. It takes the optional
            arguments ['systematic_positions']. See the documentation of
            sage.coding.linear_code_no_metric.LinearCodeSystematicEncoder
            for more details.
        """
        if not self._default_encoder_name:
            raise NotImplementedError("No encoder implemented for this code.")
        if encoder_name is None:
            encoder_name = self._default_encoder_name
        if encoder_name in self._registered_encoders:
            encClass = self._registered_encoders[encoder_name]
            try:
                return encClass(self, *args, **kwargs)
            except TypeError:
                raise ValueError(
                        "Constructing the {0} encoder failed, possibly due "
                        "to missing or incorrect parameters.\n{1}".format(
                            encoder_name, _explain_constructor(encClass)))
        else:
            raise ValueError(
                    "There is no Encoder named '{0}'. The known Encoders are: "
                    "{1}".format(encoder_name, self.encoders_available()))

    def encoders_available(self, classes=False):
        r"""
        Return a list of the available encoders' names for ``self``.

        INPUT:

        - ``classes`` -- boolean (default: ``False``); if ``classes`` is set to
          ``True``, return instead a :class:`dict` mapping available encoder
          name to the associated encoder class

        OUTPUT: list of strings, or a :class:`dict` mapping strings to classes

        EXAMPLES::

            sage: G = Matrix(GF(2), [[1,1,1,0,0,0,0], [1,0,0,1,1,0,0],
            ....:                    [0,1,0,1,0,1,0], [1,1,0,1,0,0,1]])
            sage: C = LinearCode(G)
            sage: C.encoders_available()
            ['GeneratorMatrix', 'Systematic']
            sage: dictionary = C.encoders_available(True)
            sage: sorted(dictionary.items())
            [('GeneratorMatrix', <class 'sage.coding.linear_code.LinearCodeGeneratorMatrixEncoder'>),
             ('Systematic', <class 'sage.coding.linear_code_no_metric.LinearCodeSystematicEncoder'>)]
        """
        if classes:
            return copy(self._registered_encoders)

        return sorted(self._registered_encoders)

    def unencode(self, c, encoder_name=None, nocheck=False, **kwargs):
        r"""
        Return the message corresponding to ``c``.

        This is the inverse of :meth:`encode`.

        INPUT:

        - ``c`` -- a codeword of ``self``

        - ``encoder_name`` -- (default: ``None``) name of the decoder which will be used
          to decode ``word``. The default decoder of ``self`` will be used if
          default value is kept.

        - ``nocheck`` -- boolean (default: ``False``); checks if ``c`` is in
          ``self``. You might set this to ``True`` to disable the check for
          saving computation. Note that if ``c`` is not in ``self`` and
          ``nocheck = True``, then the output of :meth:`unencode` is not
          defined (except that it will be in the message space of ``self``).

        - ``kwargs`` -- all additional arguments are forwarded to the construction of the
          encoder that is used

        OUTPUT: an element of the message space of ``encoder_name`` of ``self``

        EXAMPLES::

            sage: G = Matrix(GF(2), [[1,1,1,0,0,0,0], [1,0,0,1,1,0,0],
            ....:                    [0,1,0,1,0,1,0], [1,1,0,1,0,0,1]])
            sage: C = LinearCode(G)
            sage: c = vector(GF(2), (0, 1, 1, 0, 0, 1, 1))
            sage: C.unencode(c)
            (0, 1, 1, 0)
        """
        E = self.encoder(encoder_name, **kwargs)
        return E.unencode(c, nocheck)

    def random_element(self, *args, **kwds):
        """
        Return a random codeword; passes other positional and keyword
        arguments to ``random_element()`` method of vector space.

        OUTPUT: random element of the vector space of this code

        EXAMPLES::

            sage: C = codes.HammingCode(GF(4,'a'), 3)
            sage: C.random_element() # random test
            (1, 0, 0, a + 1, 1, a, a, a + 1, a + 1, 1, 1, 0, a + 1, a, 0, a, a, 0, a, a, 1)

        Passes extra positional or keyword arguments through::

            sage: C.random_element(prob=.5, distribution='1/n') # random test
            (1, 0, a, 0, 0, 0, 0, a + 1, 0, 0, 0, 0, 0, 0, 0, 0, a + 1, a + 1, 1, 0, 0)

        TESTS:

        Test that the codeword returned is immutable (see :issue:`16469`)::

            sage: c = C.random_element()
            sage: c.is_immutable()
            True

        Test that codeword returned has the same parent as any non-random codeword
        (see :issue:`19653`)::

            sage: C = codes.random_linear_code(GF(16, 'a'), 10, 4)
            sage: c1 = C.random_element()
            sage: c2 = C[1]
            sage: c1.parent() == c2.parent()
            True
        """
        E = self.encoder()
        M = E.message_space()
        m = M.random_element(*args, **kwds)
        c = E.encode(m)
        c.set_immutable()
        return c



# sage_setup: distribution = sagemath-flint
# distutils: libraries = flint
# distutils: depends = flint/partitions.h

################################################################################
# This file is auto-generated by the script
#   SAGE_ROOT/src/sage_setup/autogen/flint_autogen.py.
# From the commit 3e2c3a3e091106a25ca9c6fba28e02f2cbcd654a
# Do not modify by hand! Fix and rerun the script instead.
################################################################################

from libc.stdio cimport FILE
from sage.libs.gmp.types cimport *
from sage.libs.mpfr.types cimport *
from sage.libs.flint.types cimport *

cdef extern from "flint_wrap.h":
    void partitions_rademacher_bound(arf_t b, const fmpz_t n, ulong N) noexcept
    void partitions_hrr_sum_arb(arb_t x, const fmpz_t n, slong N0, slong N, int use_doubles) noexcept
    void partitions_fmpz_fmpz(fmpz_t p, const fmpz_t n, int use_doubles) noexcept
    void partitions_fmpz_ui(fmpz_t p, ulong n) noexcept
    void partitions_fmpz_ui_using_doubles(fmpz_t p, ulong n) noexcept
    void partitions_leading_fmpz(arb_t res, const fmpz_t n, slong prec) noexcept



# sage_setup: distribution = sagemath-gap
# sage.doctest: needs sage.graphs sage.groups sage.modules
"""
Weyl groups

EXAMPLES:

The Cayley graph of the Weyl Group of type ['A', 3]::

    sage: w = WeylGroup(['A',3])
    sage: d = w.cayley_graph(); d
    Digraph on 24 vertices
    sage: d.show3d(color_by_label=True, edge_size=0.01, vertex_size=0.03)               # needs sage.plot

The Cayley graph of the Weyl Group of type ['D', 4]::

    sage: w = WeylGroup(['D',4])
    sage: d = w.cayley_graph(); d
    Digraph on 192 vertices
    sage: d.show3d(color_by_label=True, edge_size=0.01, vertex_size=0.03)       # long time (less than one minute), needs sage.plot

.. TODO::

    More examples on Weyl Groups should be added here.

AUTHORS:

- Daniel Bump (2008): initial version
- Mike Hansen (2008): initial version
- Anne Schilling (2008): initial version
- Nicolas Thiéry (2008): initial version
- Volker Braun (2013): LibGAP-based matrix groups
"""
# ****************************************************************************
#       Copyright (C) 2008 Daniel Bump <bump at match.stanford.edu>,
#                          Mike Hansen <mhansen@gmail.com>
#                          Anne Schilling <anne at math.ucdavis.edu>
#                          Nicolas Thiery <nthiery at users.sf.net>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#
#                  https://www.gnu.org/licenses/
# ****************************************************************************
from sage.groups.matrix_gps.finitely_generated_gap import FinitelyGeneratedMatrixGroup_gap
from sage.groups.matrix_gps.group_element_gap import MatrixGroupElement_gap
from sage.groups.perm_gps.permgroup import PermutationGroup_generic
from sage.rings.integer_ring import ZZ
from sage.rings.rational_field import QQ
from sage.libs.gap.libgap import libgap
from sage.misc.cachefunc import cached_method
from sage.combinat.root_system.cartan_type import CartanType
from sage.combinat.root_system.cartan_matrix import CartanMatrix
from sage.combinat.root_system.reflection_group_element import RealReflectionGroupElement
from sage.matrix.constructor import matrix, diagonal_matrix
from sage.combinat.root_system.root_lattice_realizations import RootLatticeRealizations
from sage.structure.unique_representation import UniqueRepresentation
from sage.structure.richcmp import richcmp, richcmp_not_equal
from sage.categories.weyl_groups import WeylGroups
from sage.categories.finite_weyl_groups import FiniteWeylGroups
from sage.categories.affine_weyl_groups import AffineWeylGroups
from sage.categories.permutation_groups import PermutationGroups
from sage.sets.family import Family
from sage.matrix.constructor import Matrix


def WeylGroup(x, prefix=None, implementation='matrix'):
    """
    Return the Weyl group of the root system defined by the Cartan
    type (or matrix) ``ct``.

    INPUT:

    - ``x`` -- a root system or a Cartan type (or matrix)

    OPTIONAL:

    - ``prefix`` -- changes the representation of elements from matrices
      to products of simple reflections

    - ``implementation`` -- one of the following:

      * ``'matrix'`` -- as matrices acting on a root system
      * ``'permutation'`` -- as a permutation group acting on the roots

    EXAMPLES:

    The following constructions yield the same result, namely
    a weight lattice and its corresponding Weyl group::

        sage: G = WeylGroup(['F',4])
        sage: L = G.domain()

    or alternatively and equivalently::

        sage: L = RootSystem(['F',4]).ambient_space()
        sage: G = L.weyl_group()
        sage: W = WeylGroup(L)

    Either produces a weight lattice, with access to its roots and
    weights.

    ::

        sage: G = WeylGroup(['F',4])
        sage: G.order()
        1152
        sage: [s1,s2,s3,s4] = G.simple_reflections()
        sage: w = s1*s2*s3*s4; w
        [ 1/2  1/2  1/2  1/2]
        [-1/2  1/2  1/2 -1/2]
        [ 1/2  1/2 -1/2 -1/2]
        [ 1/2 -1/2  1/2 -1/2]
        sage: type(w) == G.element_class
        True
        sage: w.order()
        12
        sage: w.length() # length function on Weyl group
        4

    The default representation of Weyl group elements is as matrices.
    If you prefer, you may specify a prefix, in which case the
    elements are represented as products of simple reflections.

    ::

        sage: W=WeylGroup("C3",prefix='s')
        sage: [s1,s2,s3]=W.simple_reflections() # lets Sage parse its own output
        sage: s2*s1*s2*s3
        s1*s2*s3*s1
        sage: s2*s1*s2*s3 == s1*s2*s3*s1
        True
        sage: (s2*s3)^2==(s3*s2)^2
        True
        sage: (s1*s2*s3*s1).matrix()
        [ 0  0 -1]
        [ 0  1  0]
        [ 1  0  0]

    ::

        sage: L = G.domain()
        sage: fw = L.fundamental_weights(); fw
        Finite family {1: (1, 1, 0, 0), 2: (2, 1, 1, 0), 3: (3/2, 1/2, 1/2, 1/2), 4: (1, 0, 0, 0)}
        sage: rho = sum(fw); rho
        (11/2, 5/2, 3/2, 1/2)
        sage: w.action(rho) # action of G on weight lattice
        (5, -1, 3, 2)

    We can also do the same for arbitrary Cartan matrices::

        sage: cm = CartanMatrix([[2,-5,0],[-2,2,-1],[0,-1,2]])
        sage: W = WeylGroup(cm)
        sage: W.gens()
        (
        [-1  5  0]  [ 1  0  0]  [ 1  0  0]
        [ 0  1  0]  [ 2 -1  1]  [ 0  1  0]
        [ 0  0  1], [ 0  0  1], [ 0  1 -1]
        )
        sage: s0,s1,s2 = W.gens()
        sage: s1*s2*s1
        [ 1  0  0]
        [ 2  0 -1]
        [ 2 -1  0]
        sage: s2*s1*s2
        [ 1  0  0]
        [ 2  0 -1]
        [ 2 -1  0]
        sage: s0*s1*s0*s2*s0
        [ 9  0 -5]
        [ 2  0 -1]
        [ 0  1 -1]

    Same Cartan matrix, but with a prefix to display using simple reflections::

        sage: W = WeylGroup(cm, prefix='s')
        sage: s0,s1,s2 = W.gens()
        sage: s0*s2*s1
        s2*s0*s1
        sage: (s1*s2)^3
        1
        sage: (s0*s1)^5
        s0*s1*s0*s1*s0*s1*s0*s1*s0*s1
        sage: s0*s1*s2*s1*s2
        s2*s0*s1
        sage: s0*s1*s2*s0*s2
        s0*s1*s0

    TESTS::

        sage: TestSuite(WeylGroup(["A",3])).run()
        sage: TestSuite(WeylGroup(["A",2,1])).run() # long time

        sage: W = WeylGroup(['A',3,1])
        sage: s = W.simple_reflections()
        sage: w = s[0]*s[1]*s[2]
        sage: w.reduced_word()
        [0, 1, 2]
        sage: w = s[0]*s[2]
        sage: w.reduced_word()
        [2, 0]
        sage: W = groups.misc.WeylGroup(['A',3,1])
    """
    if implementation == "permutation":
        return WeylGroup_permutation(x, prefix)
    if implementation != "matrix":
        raise ValueError("invalid implementation")

    if x in RootLatticeRealizations:
        return WeylGroup_gens(x, prefix=prefix)

    try:
        ct = CartanType(x)
    except TypeError:
        ct = CartanMatrix(x)  # See if it is a Cartan matrix
    if ct.is_finite():
        return WeylGroup_gens(ct.root_system().ambient_space(), prefix=prefix)
    return WeylGroup_gens(ct.root_system().root_space(), prefix=prefix)


class WeylGroup_gens(UniqueRepresentation,
                     FinitelyGeneratedMatrixGroup_gap):

    @staticmethod
    def __classcall__(cls, domain, prefix=None):
        return super().__classcall__(cls, domain, prefix)

    def __init__(self, domain, prefix):
        """
        EXAMPLES::

            sage: G = WeylGroup(['B',3])
            sage: TestSuite(G).run()                                                    # needs sage.rings.number_field

            sage: cm = CartanMatrix([[2,-5,0],[-2,2,-1],[0,-1,2]])
            sage: W = WeylGroup(cm)
            sage: TestSuite(W).run() # long time

        TESTS::

            sage: W = WeylGroup(SymmetricGroup(1))
        """
        self._domain = domain
        if self.cartan_type().is_affine():
            category = AffineWeylGroups()
        elif self.cartan_type().is_finite():
            category = FiniteWeylGroups()
        else:
            category = WeylGroups()
        if self.cartan_type().is_irreducible():
            category = category.Irreducible()
        self.n = domain.dimension()  # Really needed?
        self._prefix = prefix

        # FinitelyGeneratedMatrixGroup_gap takes plain matrices as input
        gens_matrix = [self.morphism_matrix(self.domain().simple_reflection(i))
                       for i in self.index_set()]
        if not gens_matrix:
            libgap_group = libgap.Group([], matrix(ZZ, 1, 1, [1]))
        else:
            libgap_group = libgap.Group(gens_matrix)
        degree = ZZ(self.domain().dimension())
        ring = self.domain().base_ring()
        FinitelyGeneratedMatrixGroup_gap.__init__(
            self, degree, ring, libgap_group, category=category)

    @cached_method
    def cartan_type(self):
        """
        Return the ``CartanType`` associated to ``self``.

        EXAMPLES::

            sage: G = WeylGroup(['F',4])
            sage: G.cartan_type()
            ['F', 4]
        """
        return self.domain().cartan_type()

    @cached_method
    def index_set(self):
        """
        Return the index set of ``self``.

        EXAMPLES::

            sage: G = WeylGroup(['F',4])
            sage: G.index_set()
            (1, 2, 3, 4)
            sage: G = WeylGroup(['A',3,1])
            sage: G.index_set()
            (0, 1, 2, 3)
        """
        return self.cartan_type().index_set()

    # Should be implemented in (morphisms of) modules with basis
    def morphism_matrix(self, f):
        return matrix(self.domain().base_ring(),
                      [f(b).to_vector()
                       for b in self.domain().basis()]).transpose()

    def from_morphism(self, f):
        return self._element_constructor_(self.morphism_matrix(f))

    @cached_method
    def simple_reflections(self):
        """
        Return the simple reflections of ``self``, as a family.

        EXAMPLES:

        There are the simple reflections for the symmetric group::

            sage: W=WeylGroup(['A',2])
            sage: s = W.simple_reflections(); s
            Finite family {1: [0 1 0]
            [1 0 0]
            [0 0 1], 2: [1 0 0]
            [0 0 1]
            [0 1 0]}

        As a special feature, for finite irreducible root systems,
        s[0] gives the reflection along the highest root::

            sage: s[0]
            [0 0 1]
            [0 1 0]
            [1 0 0]

        We now look at some further examples::

            sage: W=WeylGroup(['A',2,1])
            sage: W.simple_reflections()
            Finite family {0: [-1  1  1]
            [ 0  1  0]
            [ 0  0  1], 1: [ 1  0  0]
            [ 1 -1  1]
            [ 0  0  1], 2: [ 1  0  0]
            [ 0  1  0]
            [ 1  1 -1]}
            sage: W = WeylGroup(['F',4])
            sage: [s1,s2,s3,s4] = W.simple_reflections()
            sage: w = s1*s2*s3*s4; w
            [ 1/2  1/2  1/2  1/2]
            [-1/2  1/2  1/2 -1/2]
            [ 1/2  1/2 -1/2 -1/2]
            [ 1/2 -1/2  1/2 -1/2]
            sage: s4^2 == W.one()
            True
            sage: type(w) == W.element_class
            True
        """
        return self.domain().simple_reflections().map(self.from_morphism)

    def reflections(self):
        """
        Return the reflections of ``self``.

        The reflections of a Coxeter group `W` are the conjugates of
        the simple reflections. They are in bijection with the positive
        roots, for given a positive root, we may have the reflection in
        the hyperplane orthogonal to it. This method returns a family
        indexed by the positive roots taking values in the reflections.
        This requires ``self`` to be a finite Weyl group.

        .. NOTE::

            Prior to :issue:`20027`, the reflections were the keys
            of the family and the values were the positive roots.

        EXAMPLES::

            sage: W = WeylGroup("B2", prefix='s')
            sage: refdict = W.reflections(); refdict
            Finite family {(1, -1): s1, (0, 1): s2, (1, 1): s2*s1*s2, (1, 0): s1*s2*s1}
            sage: [r+refdict[r].action(r) for r in refdict.keys()]
            [(0, 0), (0, 0), (0, 0), (0, 0)]

            sage: W = WeylGroup(['A',2,1], prefix='s')
            sage: W.reflections()
            Lazy family (real root to reflection(i))_{i in
                        Positive real roots of type ['A', 2, 1]}

        TESTS::

            sage: CM = CartanMatrix([[2,-6],[-1,2]])
            sage: W = WeylGroup(CM, prefix='s')
            sage: W.reflections()
            Traceback (most recent call last):
            ...
            NotImplementedError: only implemented for finite and affine Cartan types
        """
        prr = self.domain().positive_real_roots()

        def to_elt(alp):
            ref = self.domain().reflection(alp)
            m = Matrix([ref(x).to_vector() for x in self.domain().basis()])
            return self(m.transpose())
        return Family(prr, to_elt, name="real root to reflection")

    def _repr_(self):
        """
        EXAMPLES::

            sage: WeylGroup(['A', 1])
            Weyl Group of type ['A', 1] (as a matrix group acting on the ambient space)
            sage: WeylGroup(['A', 3, 1])
            Weyl Group of type ['A', 3, 1] (as a matrix group acting on the root space)
        """
        domain = self._domain._name_string(capitalize=False,
                                           base_ring=False,
                                           type=False)
        return "Weyl Group of type %s (as a matrix group acting on the %s)" % (self.cartan_type(), domain)

    def character_table(self):
        """
        Return the character table as a matrix.

        Each row is an irreducible character. For larger tables you
        may preface this with a command such as
        gap.eval("SizeScreen([120,40])") in order to widen the screen.

        EXAMPLES::

            sage: WeylGroup(['A',3]).character_table() # random
            CT1
            <BLANKLINE>
                 2  3  2  2  .  3
                 3  1  .  .  1  .
            <BLANKLINE>
                   1a 4a 2a 3a 2b
            <BLANKLINE>
            X.1     1 -1 -1  1  1
            X.2     3  1 -1  . -1
            X.3     2  .  . -1  2
            X.4     3 -1  1  . -1
            X.5     1  1  1  1  1
        """
        G = libgap.Group([libgap(g) for g in self.gens()])
        ctbl = libgap.CharacterTable(G)
        return ctbl.Display()

    @cached_method
    def one(self):
        """
        Return the unit element of the Weyl group.

        EXAMPLES::

            sage: W = WeylGroup(['A',3])
            sage: e = W.one(); e
            [1 0 0 0]
            [0 1 0 0]
            [0 0 1 0]
            [0 0 0 1]
            sage: type(e) == W.element_class
            True
        """
        return self._element_constructor_(matrix(QQ, self.n, self.n, 1))

    unit = one  # For backward compatibility

    def domain(self):
        """
        Return the domain of the element of ``self``, that is the
        root lattice realization on which they act.

        EXAMPLES::

            sage: G = WeylGroup(['F',4])
            sage: G.domain()
            Ambient space of the Root system of type ['F', 4]
            sage: G = WeylGroup(['A',3,1])
            sage: G.domain()
            Root space over the Rational Field of the Root system of type ['A', 3, 1]
        """
        return self._domain

    def simple_reflection(self, i):
        """
        Return the `i`-th simple reflection.

        EXAMPLES::

            sage: G = WeylGroup(['F',4])
            sage: G.simple_reflection(1)
            [1 0 0 0]
            [0 0 1 0]
            [0 1 0 0]
            [0 0 0 1]
            sage: W=WeylGroup(['A',2,1])
            sage: W.simple_reflection(1)
            [ 1  0  0]
            [ 1 -1  1]
            [ 0  0  1]
        """
        if i not in self.index_set():
            raise ValueError("i must be in the index set")
        return self.simple_reflections()[i]

    def long_element_hardcoded(self):
        """
        Return the long Weyl group element (hardcoded data).

        Do we really want to keep it? There is a generic
        implementation which works in all cases. The hardcoded should
        have a better complexity (for large classical types), but
        there is a cache, so does this really matter?

        EXAMPLES::

            sage: types = [ ['A',5],['B',3],['C',3],['D',4],['G',2],['F',4],['E',6] ]
            sage: [WeylGroup(t).long_element().length() for t in types]
            [15, 9, 9, 12, 6, 24, 36]
            sage: all(WeylGroup(t).long_element() == WeylGroup(t).long_element_hardcoded() for t in types)  # long time (17s on sage.math, 2011)
            True
        """
        typ = self.cartan_type()
        if typ[0] == 'D' and typ[1] % 2:
            l = [-1 for i in range(self.n - 1)]
            l.append(1)
            m = diagonal_matrix(QQ, l)
        elif typ[0] == 'A':
            l = [0 for k in range((self.n)**2)]
            for k in range(self.n - 1, (self.n)**2 - 1, self.n - 1):
                l[k] = 1
            m = matrix(QQ, self.n, l)
        elif typ[0] == 'E':
            if typ[1] == 6:
                half = QQ((1, 2))
                l = [[-half, -half, -half, half, 0, 0, 0, 0],
                     [-half, -half, half, -half, 0, 0, 0, 0],
                     [-half, half, -half, -half, 0, 0, 0, 0],
                     [half, -half, -half, -half, 0, 0, 0, 0],
                     [0, 0, 0, 0, half, half, half, -half],
                     [0, 0, 0, 0, half, half, -half, half],
                     [0, 0, 0, 0, half, -half, half, half],
                     [0, 0, 0, 0, -half, half, half, half]]
                m = matrix(QQ, 8, l)
            else:
                raise NotImplementedError("not implemented yet for this type")
        elif typ[0] == 'G':
            third = QQ((1, 3))
            twothirds = QQ((2, 3))
            l = [[-third, twothirds, twothirds],
                 [twothirds, -third, twothirds],
                 [twothirds, twothirds, -third]]
            m = matrix(QQ, 3, l)
        else:
            m = diagonal_matrix([-1] * self.n)
        return self(m)

    def classical(self):
        """
        If ``self`` is a Weyl group from an affine Cartan Type, this give
        the classical parabolic subgroup of ``self``.

        Caveat: we assume that 0 is a special node of the Dynkin diagram

        .. TODO:: extract parabolic subgroup method

        EXAMPLES::

            sage: G = WeylGroup(['A',3,1])
            sage: G.classical()
            Parabolic Subgroup of the Weyl Group of type ['A', 3, 1]
             (as a matrix group acting on the root space)
            sage: WeylGroup(['A',3]).classical()
            Traceback (most recent call last):
            ...
            ValueError: classical subgroup only defined for affine types
        """
        if not self.cartan_type().is_affine():
            raise ValueError("classical subgroup only defined for affine types")
        return ClassicalWeylSubgroup(self._domain, prefix=self._prefix)


class ClassicalWeylSubgroup(WeylGroup_gens):
    """
    A class for Classical Weyl Subgroup of an affine Weyl Group.

    EXAMPLES::

        sage: G = WeylGroup(["A",3,1]).classical()
        sage: G
        Parabolic Subgroup of the Weyl Group of type ['A', 3, 1] (as a matrix group acting on the root space)
        sage: G.category()
        Category of finite irreducible Weyl groups
        sage: G.cardinality()
        24
        sage: G.index_set()
        (1, 2, 3)
        sage: TestSuite(G).run()                                                        # needs sage.rings.number_field

    TESTS::

        sage: from sage.combinat.root_system.weyl_group import ClassicalWeylSubgroup
        sage: H = ClassicalWeylSubgroup(RootSystem(["A", 3, 1]).root_space(), prefix=None)
        sage: H is G
        True

    Caveat: the interface is likely to change. The current main
    application is for plots.

    .. TODO::

        implement:

        - Parabolic subrootsystems
        - Parabolic subgroups with a set of nodes as argument
    """
    @cached_method
    def cartan_type(self):
        """
        EXAMPLES::

            sage: WeylGroup(['A',3,1]).classical().cartan_type()
            ['A', 3]
            sage: WeylGroup(['A',3,1]).classical().index_set()
            (1, 2, 3)

        Note: won't be needed, once the lattice will be a parabolic sub root system
        """
        return self.domain().cartan_type().classical()

    def simple_reflections(self):
        """
        EXAMPLES::

            sage: WeylGroup(['A',2,1]).classical().simple_reflections()
            Finite family {1: [ 1  0  0]
                              [ 1 -1  1]
                              [ 0  0  1],
                           2: [ 1  0  0]
                              [ 0  1  0]
                              [ 1  1 -1]}

        Note: won't be needed, once the lattice will be a parabolic sub root system
        """
        return Family({i: self.from_morphism(self.domain().simple_reflection(i))
                       for i in self.index_set()})

    def __repr__(self):
        """
        EXAMPLES::

            sage: WeylGroup(['A',2,1]).classical()
            Parabolic Subgroup of the Weyl Group of type ['A', 2, 1] (as a matrix group acting on the root space)
            sage: WeylGroup(['C',4,1]).classical()
            Parabolic Subgroup of the Weyl Group of type ['C', 4, 1] (as a matrix group acting on the root space)
            sage: RootSystem(['C',3,1]).coweight_lattice().weyl_group().classical()
            Parabolic Subgroup of the Weyl Group of type ['C', 3, 1]^* (as a matrix group acting on the coweight lattice)
            sage: RootSystem(['C',4,1]).coweight_lattice().weyl_group().classical()
            Parabolic Subgroup of the Weyl Group of type ['C', 4, 1]^* (as a matrix group acting on the coweight lattice)
        """
        domain = self._domain._name_string(capitalize=False,
                                           base_ring=False,
                                           type=False)
        return "Parabolic Subgroup of the Weyl Group of type %s (as a matrix group acting on the %s)" % (self.domain().cartan_type(), domain)

    def weyl_group(self, prefix='hereditary'):
        """
        Return the Weyl group associated to the parabolic subgroup.

        EXAMPLES::

            sage: WeylGroup(['A',4,1]).classical().weyl_group()
            Weyl Group of type ['A', 4, 1] (as a matrix group acting on the root space)
            sage: WeylGroup(['C',4,1]).classical().weyl_group()
            Weyl Group of type ['C', 4, 1] (as a matrix group acting on the root space)
            sage: WeylGroup(['E',8,1]).classical().weyl_group()
            Weyl Group of type ['E', 8, 1] (as a matrix group acting on the root space)
        """
        if prefix == "hereditary":
            prefix = self._prefix
        return self.domain().weyl_group(prefix)

    def _test_is_finite(self, **options):
        """
        Test some internal invariants.

        EXAMPLES::

            sage: WeylGroup(['A', 2, 1]).classical()._test_is_finite()
            sage: WeylGroup(['B', 3, 1]).classical()._test_is_finite()
        """
        tester = self._tester(**options)
        tester.assertTrue(not self.weyl_group(self._prefix).is_finite())
        tester.assertTrue(self.is_finite())


class WeylGroupElement(MatrixGroupElement_gap):
    """
    Class for a Weyl Group elements
    """

    def __init__(self, parent, g, check=False):
        """
        EXAMPLES::

            sage: G = WeylGroup(['A',2])
            sage: s1 = G.simple_reflection(1)
            sage: TestSuite(s1).run()
        """
        MatrixGroupElement_gap.__init__(self, parent, g, check=check)
        self._parent = parent

    def __hash__(self):
        return hash(self.matrix())

    def to_matrix(self):
        """
        Return ``self`` as a matrix.

        EXAMPLES::

            sage: G = WeylGroup(['A',2])
            sage: s1 = G.simple_reflection(1)
            sage: s1.to_matrix() == s1.matrix()
            True
        """
        return self.matrix()

    def domain(self):
        """
        Return the ambient lattice associated with ``self``.

        EXAMPLES::

            sage: W = WeylGroup(['A',2])
            sage: s1 = W.simple_reflection(1)
            sage: s1.domain()
            Ambient space of the Root system of type ['A', 2]
        """
        return self._parent.domain()

    def _repr_(self):
        """
        EXAMPLES::

            sage: W = WeylGroup(['A',2,1], prefix='s')
            sage: [s0,s1,s2] = W.simple_reflections()
            sage: s0*s1
            s0*s1
            sage: W = WeylGroup(['A',2,1])
            sage: [s0,s1,s2]=W.simple_reflections()
            sage: s0*s1
            [ 0 -1  2]
            [ 1 -1  1]
            [ 0  0  1]
        """
        if self._parent._prefix is None:
            return MatrixGroupElement_gap._repr_(self)

        redword = self.reduced_word()
        if len(redword) == 0:
            return "1"

        ret = "".join("%s%d*" % (self._parent._prefix, i)
                      for i in redword[:-1])
        return ret + "%s%d" % (self._parent._prefix, redword[-1])

    def _latex_(self):
        r"""
        Return the latex representation of ``self``.

        EXAMPLES::

            sage: W = WeylGroup(['A',2,1], prefix='s')
            sage: [s0,s1,s2] = W.simple_reflections()
            sage: latex(s0*s1)  # indirect doctest
            s_{0}s_{1}
            sage: W = WeylGroup(['A',2,1])
            sage: [s0,s1,s2] = W.simple_reflections()
            sage: latex(s0*s1)
            \left(\begin{array}{rrr}
            0 & -1 & 2 \\
            1 & -1 & 1 \\
            0 & 0 & 1
            \end{array}\right)
        """
        if self._parent._prefix is None:
            return MatrixGroupElement_gap._latex_(self)

        redword = self.reduced_word()
        if not redword:
            return "1"

        return "".join("%s_{%d}" % (self._parent._prefix, i) for i in redword)

    def __eq__(self, other):
        """
        EXAMPLES::

            sage: W = WeylGroup(['A',3])
            sage: s = W.simple_reflections()
            sage: s[1] == s[1]
            True
            sage: s[1] == s[2]
            False

        Note: this implementation of :meth:`__eq__` is not much faster
        than :meth:`__cmp__`. But it turned out to be useful for
        subclasses overriding __cmp__ with something slow for specific
        purposes.
        """
        return (self.__class__ == other.__class__ and
                self._parent == other._parent and
                self.matrix() == other.matrix())

    def _richcmp_(self, other, op):
        """
        EXAMPLES::

            sage: W = WeylGroup(['A',3])
            sage: s = W.simple_reflections()
            sage: s[1] == s[1]
            True
            sage: s[1] == s[2]
            False
        """
        if self._parent.cartan_type() != other._parent.cartan_type():
            return richcmp_not_equal(self._parent.cartan_type(),
                                     other._parent.cartan_type(), op)
        return richcmp(self.matrix(), other.matrix(), op)

    def action(self, v):
        """
        Return the action of ``self`` on the vector `v`.

        EXAMPLES::

            sage: W = WeylGroup(['A',2])
            sage: s = W.simple_reflections()
            sage: v = W.domain()([1,0,0])
            sage: s[1].action(v)
            (0, 1, 0)

            sage: W = WeylGroup(RootSystem(['A',2]).root_lattice())
            sage: s = W.simple_reflections()
            sage: alpha = W.domain().simple_roots()
            sage: s[1].action(alpha[1])
            -alpha[1]

            sage: W=WeylGroup(['A',2,1])
            sage: alpha = W.domain().simple_roots()
            sage: s = W.simple_reflections()
            sage: s[1].action(alpha[1])
            -alpha[1]
            sage: s[1].action(alpha[0])
            alpha[0] + alpha[1]
        """
        if v not in self.domain():
            raise ValueError(f"{v} is not in the domain")
        return self.domain().from_vector(self.matrix() * v.to_vector())

    # #######################################################################
    # Descents
    # #######################################################################

    def has_descent(self, i, positive=False, side='right') -> bool:
        """
        Test if ``self`` has a descent at position ``i``.

        An element `w` has a descent in position `i` if `w` is
        on the strict negative side of the `i`-th simple reflection
        hyperplane.

        If ``positive`` is ``True``, tests if it is on the strict
        positive side instead.

        EXAMPLES::

            sage: W = WeylGroup(['A',3])
            sage: s = W.simple_reflections()
            sage: [W.one().has_descent(i) for i in W.domain().index_set()]
            [False, False, False]
            sage: [s[1].has_descent(i) for i in W.domain().index_set()]
            [True, False, False]
            sage: [s[2].has_descent(i) for i in W.domain().index_set()]
            [False, True, False]
            sage: [s[3].has_descent(i) for i in W.domain().index_set()]
            [False, False, True]
            sage: [s[3].has_descent(i, True) for i in W.domain().index_set()]
            [True, True, False]
            sage: W = WeylGroup(['A',3,1])
            sage: s = W.simple_reflections()
            sage: [W.one().has_descent(i) for i in W.domain().index_set()]
            [False, False, False, False]
            sage: [s[0].has_descent(i) for i in W.domain().index_set()]
            [True, False, False, False]
            sage: w = s[0] * s[1]
            sage: [w.has_descent(i) for i in W.domain().index_set()]
            [False, True, False, False]
            sage: [w.has_descent(i, side = "left") for i in W.domain().index_set()]
            [True, False, False, False]
            sage: w = s[0] * s[2]
            sage: [w.has_descent(i) for i in W.domain().index_set()]
            [True, False, True, False]
            sage: [w.has_descent(i, side = "left") for i in W.domain().index_set()]
            [True, False, True, False]

            sage: W = WeylGroup(['A',3])
            sage: W.one().has_descent(0)
            True
            sage: W.w0.has_descent(0)
            False
        """
#        s=self.parent().lattice().rho().scalar(self.action(self.parent().lattice().simple_root(i)))
#        if positive:
#            return s > 0
#        else:
#            return s < 0
        L = self.domain()
        # Choose the method depending on the side and the availability of rho and is_positive_root
        if not hasattr(L.element_class, "is_positive_root"):
            use_rho = True
        elif not hasattr(L, "rho"):
            use_rho = False
        else:
            use_rho = side == "left"

        element = self if use_rho is (side == "left") else ~self

        if use_rho:
            s = element.action(L.rho()).scalar(L.alphacheck()[i]) >= 0
        else:
            s = element.action(L.alpha()[i]).is_positive_root()

        return s is positive

    def has_left_descent(self, i) -> bool:
        """
        Test if ``self`` has a left descent at position ``i``.

        EXAMPLES::

            sage: W = WeylGroup(['A',3])
            sage: s = W.simple_reflections()
            sage: [W.one().has_left_descent(i) for i in W.domain().index_set()]
            [False, False, False]
            sage: [s[1].has_left_descent(i) for i in W.domain().index_set()]
            [True, False, False]
            sage: [s[2].has_left_descent(i) for i in W.domain().index_set()]
            [False, True, False]
            sage: [s[3].has_left_descent(i) for i in W.domain().index_set()]
            [False, False, True]
            sage: [(s[3]*s[2]).has_left_descent(i) for i in W.domain().index_set()]
            [False, False, True]
        """
        return self.has_descent(i, side='left')

    def has_right_descent(self, i) -> bool:
        """
        Test if ``self`` has a right descent at position ``i``.

        EXAMPLES::

            sage: W = WeylGroup(['A',3])
            sage: s = W.simple_reflections()
            sage: [W.one().has_right_descent(i) for i in W.domain().index_set()]
            [False, False, False]
            sage: [s[1].has_right_descent(i) for i in W.domain().index_set()]
            [True, False, False]
            sage: [s[2].has_right_descent(i) for i in W.domain().index_set()]
            [False, True, False]
            sage: [s[3].has_right_descent(i) for i in W.domain().index_set()]
            [False, False, True]
            sage: [(s[3]*s[2]).has_right_descent(i) for i in W.domain().index_set()]
            [False, True, False]
        """
        return self.has_descent(i, side='right')

    def apply_simple_reflection(self, i, side='right'):
        s = self.parent().simple_reflections()
        if side == "right":
            return self * s[i]
        return s[i] * self

    def to_permutation(self):
        """
        A first approximation of to_permutation ...

        This assumes types A,B,C,D on the ambient lattice

        This further assume that the basis is indexed by 0,1,...
        and returns a permutation of (5,4,2,3,1) (beuargl), as a tuple
        """
        W = self.parent()
        e = W.domain().basis()
        return tuple(c * (j + 1) for i in e.keys()
                     for (j, c) in self.action(e[i]))

    def to_permutation_string(self):
        """
        EXAMPLES::

            sage: W = WeylGroup(["A",3])
            sage: s = W.simple_reflections()
            sage: (s[1]*s[2]*s[3]).to_permutation_string()
            '2341'
        """
        return "".join(str(i) for i in self.to_permutation())


WeylGroup_gens.Element = WeylGroupElement


class WeylGroup_permutation(UniqueRepresentation, PermutationGroup_generic):
    """
    A Weyl group given as a permutation group.
    """
    @staticmethod
    def __classcall__(cls, cartan_type, prefix=None):
        """
        Normalize input to ensure a unique representation.

        EXAMPLES::

            sage: W1 = WeylGroup(['B',2], implementation='permutation')
            sage: W2 = WeylGroup(CartanType(['B',2]), implementation='permutation')
            sage: W1 is W2
            True
        """
        return super().__classcall__(cls, CartanType(cartan_type), prefix)

    def __init__(self, cartan_type, prefix):
        """
        Initialize ``self``.

        EXAMPLES::

            sage: W = WeylGroup(['F',4], implementation='permutation')
            sage: TestSuite(W).run()                                                    # needs sage.rings.number_field
        """
        self._cartan_type = cartan_type
        self._index_set = cartan_type.index_set()
        self._index_set_inverse = {ii: i for i, ii in enumerate(cartan_type.index_set())}
        self._reflection_representation = None
        self._prefix = prefix
        Q = cartan_type.root_system().root_lattice()
        Phi = list(Q.positive_roots()) + [-x for x in Q.positive_roots()]
        p = [[Phi.index(x.weyl_action([i])) + 1 for x in Phi]
             for i in self._cartan_type.index_set()]
        cat = FiniteWeylGroups()
        if self._cartan_type.is_irreducible():
            cat = cat.Irreducible()
        cat = (cat, PermutationGroups().Finite())
        PermutationGroup_generic.__init__(self, gens=p, canonicalize=False, category=cat)

    def iteration(self, algorithm='breadth', tracking_words=True):
        r"""
        Return an iterator going through all elements in ``self``.

        INPUT:

        - ``algorithm`` -- (default: ``'breadth'``) must be one of
          the following:

          * ``'breadth'`` -- iterate over in a linear extension of the
            weak order
          * ``'depth'`` -- iterate by a depth-first-search

        - ``tracking_words`` -- boolean (default: ``True``); whether or not to keep
          track of the reduced words and store them in ``_reduced_word``

        .. NOTE::

            The fastest iteration is the depth first algorithm without
            tracking words. In particular, ``'depth'`` is ~1.5x faster.

        EXAMPLES::

            sage: W = WeylGroup(["B",2], implementation='permutation')

            sage: for w in W.iteration("breadth",True):
            ....:     print("%s %s"%(w, w._reduced_word))
            () []
            (1,3)(2,6)(5,7) [1]
            (1,5)(2,4)(6,8) [0]
            (1,7,5,3)(2,4,6,8) [0, 1]
            (1,3,5,7)(2,8,6,4) [1, 0]
            (2,8)(3,7)(4,6) [1, 0, 1]
            (1,7)(3,5)(4,8) [0, 1, 0]
            (1,5)(2,6)(3,7)(4,8) [0, 1, 0, 1]

            sage: for w in W.iteration("depth", False): w
            ()
            (1,3)(2,6)(5,7)
            (1,5)(2,4)(6,8)
            (1,3,5,7)(2,8,6,4)
            (1,7)(3,5)(4,8)
            (1,7,5,3)(2,4,6,8)
            (2,8)(3,7)(4,6)
            (1,5)(2,6)(3,7)(4,8)

        TESTS::

            sage: W = WeylGroup(["A",0], implementation='permutation')
            sage: list(W)
            [()]
            sage: W[0]
            ()
        """
        from sage.combinat.root_system.reflection_group_c import Iterator
        if self.rank() == 0:
            return iter([self.one()])
        return iter(Iterator(self, N=self.number_of_reflections(),
                             algorithm=algorithm, tracking_words=tracking_words))

    def __iter__(self):
        r"""
        Return an iterator going through all elements in ``self``.

        For options and faster iteration see :meth:`iteration`.

        EXAMPLES::

            sage: W = WeylGroup(["B",2], implementation='permutation')
            sage: for w in W: print("%s %s"%(w, w._reduced_word))
            () []
            (1,3)(2,6)(5,7) [1]
            (1,5)(2,4)(6,8) [0]
            (1,7,5,3)(2,4,6,8) [0, 1]
            (1,3,5,7)(2,8,6,4) [1, 0]
            (2,8)(3,7)(4,6) [1, 0, 1]
            (1,7)(3,5)(4,8) [0, 1, 0]
            (1,5)(2,6)(3,7)(4,8) [0, 1, 0, 1]
        """
        return self.iteration(algorithm='breadth', tracking_words=True)

    def _coerce_map_from_(self, P):
        """
        Return ``True`` if ``P`` is a Weyl group of the same
        Cartan type and ``False`` otherwise.

        EXAMPLES::

            sage: W = WeylGroup(["B",4], implementation='permutation')
            sage: W2 = WeylGroup(["B",4])
            sage: W._coerce_map_from_(W2)
            True
            sage: W3 = WeylGroup(["B",5])
            sage: W.has_coerce_map_from(W3)
            False
            sage: W4 = CoxeterGroup(["B",4])                                            # needs sage.rings.number_field
            sage: W.has_coerce_map_from(W4)                                             # needs sage.rings.number_field
            False
            sage: W5 = WeylGroup(["C",4], implementation='permutation')
            sage: W.has_coerce_map_from(W5)
            False
        """
        return isinstance(P, WeylGroup_gens) and P.cartan_type() is self.cartan_type()

    @cached_method
    def rank(self):
        """
        Return the rank of ``self``.

        EXAMPLES::

            sage: W = WeylGroup(['A',4], implementation='permutation')
            sage: W.rank()
            4
        """
        return self._cartan_type.rank()

    def simple_reflection(self, i):
        r"""
        Return the ``i``-th simple reflection of ``self``.

        EXAMPLES::

            sage: W = WeylGroup(['A',4], implementation='permutation')
            sage: W.simple_reflection(1)
            (1,11)(2,5)(6,8)(9,10)(12,15)(16,18)(19,20)
            sage: W.simple_reflections()
            Finite family {1: (1,11)(2,5)(6,8)(9,10)(12,15)(16,18)(19,20),
                           2: (1,5)(2,12)(3,6)(7,9)(11,15)(13,16)(17,19),
                           3: (2,6)(3,13)(4,7)(5,8)(12,16)(14,17)(15,18),
                           4: (3,7)(4,14)(6,9)(8,10)(13,17)(16,19)(18,20)}
        """
        return self.gens()[self._index_set_inverse[i]]

    @cached_method
    def simple_roots(self):
        """
        Return the simple roots of ``self``.

        EXAMPLES::

            sage: W = WeylGroup(['A',4], implementation='permutation')
            sage: W.simple_roots()
            Finite family {1: (1, 0, 0, 0), 2: (0, 1, 0, 0),
                           3: (0, 0, 1, 0), 4: (0, 0, 0, 1)}
        """
        Q = self._cartan_type.root_system().root_lattice()
        roots = [al.to_vector() for al in Q.simple_roots()]
        for v in roots:
            v.set_immutable()
        return Family(self._index_set, lambda i: roots[self._index_set_inverse[i]])

    independent_roots = simple_roots

    @cached_method
    def index_set(self):
        """
        Return the index set of ``self``.

        EXAMPLES::

            sage: W = WeylGroup(['A',4], implementation='permutation')
            sage: W.index_set()
            (1, 2, 3, 4)
        """
        return self._index_set

    @cached_method
    def reflection_index_set(self):
        """
        Return the index set of reflections of ``self``.

        EXAMPLES::

            sage: W = WeylGroup(['A',3], implementation='permutation')
            sage: W.reflection_index_set()
            (1, 2, 3, 4, 5, 6)
        """
        return tuple(range(1, self.number_of_reflections() + 1))

    def cartan_type(self):
        """
        Return the Cartan type of ``self``.

        EXAMPLES::

            sage: W = WeylGroup(['A',4], implementation='permutation')
            sage: W.cartan_type()
            ['A', 4]
        """
        return self._cartan_type

    @cached_method
    def roots(self):
        """
        Return the roots of ``self``.

        EXAMPLES::

            sage: W = WeylGroup(['G',2], implementation='permutation')
            sage: W.roots()
            ((1, 0),
             (0, 1),
             (1, 1),
             (3, 1),
             (2, 1),
             (3, 2),
             (-1, 0),
             (0, -1),
             (-1, -1),
             (-3, -1),
             (-2, -1),
             (-3, -2))
        """
        Q = self._cartan_type.root_system().root_lattice()
        roots = ([x.to_vector() for x in Q.positive_roots()]
                 + [-x.to_vector() for x in Q.positive_roots()])
        for v in roots:
            v.set_immutable()
        return tuple(roots)

    def positive_roots(self):
        """
        Return the positive roots of ``self``.

        EXAMPLES::

            sage: W = WeylGroup(['C',3], implementation='permutation')
            sage: W.positive_roots()
            ((1, 0, 0),
             (0, 1, 0),
             (0, 0, 1),
             (1, 1, 0),
             (0, 1, 1),
             (0, 2, 1),
             (1, 1, 1),
             (2, 2, 1),
             (1, 2, 1))
        """
        return self.roots()[:self.number_of_reflections()]

    @cached_method
    def number_of_reflections(self):
        """
        Return the number of reflections in ``self``.

        EXAMPLES::

            sage: W = WeylGroup(['D',4], implementation='permutation')
            sage: W.number_of_reflections()
            12
        """
        return len(list(self._cartan_type.root_system().root_lattice().positive_roots()))

    @cached_method
    def distinguished_reflections(self):
        """
        Return the reflections of ``self``.

        EXAMPLES::

            sage: W = WeylGroup(['B',2], implementation='permutation')
            sage: W.distinguished_reflections()
            Finite family {1: (1,5)(2,4)(6,8), 2: (1,3)(2,6)(5,7),
                           3: (2,8)(3,7)(4,6), 4: (1,7)(3,5)(4,8)}
        """
        Q = self._cartan_type.root_system().root_lattice()
        pos_roots = list(Q.positive_roots())
        Phi = pos_roots + [-x for x in pos_roots]

        def build_elt(index):
            r = pos_roots[index]
            perm = [Phi.index(x.reflection(r)) + 1 for x in Phi]
            return self.element_class(perm, self, check=False)
        return Family(self.reflection_index_set(), lambda i: build_elt(i - 1))

    reflections = distinguished_reflections

    def simple_root_index(self, i):
        r"""
        Return the index of the simple root `\alpha_i`.

        This is the position of `\alpha_i` in the list of simple roots.

        EXAMPLES::

            sage: W = WeylGroup(['A',3], implementation='permutation')
            sage: [W.simple_root_index(i) for i in W.index_set()]
            [0, 1, 2]
        """
        return self._index_set_inverse[i]

    class Element(RealReflectionGroupElement):
        def _repr_(self):
            """
            EXAMPLES::

                sage: W = WeylGroup(['A',3], prefix='s', implementation='permutation')
                sage: [s1,s2,s3] = W.simple_reflections()
                sage: s1*s2
                s1*s2
                sage: W = WeylGroup(['A',3], implementation='permutation')
                sage: [s1,s2,s3] = W.simple_reflections()
                sage: s1*s2
                (1,10,2)(3,5,6)(4,8,7)(9,11,12)
            """
            if self.parent()._prefix is None:
                return RealReflectionGroupElement._repr_(self)
            redword = self.reduced_word()
            if not redword:
                return "1"
            return "*".join("%s%d" % (self.parent()._prefix, i)
                            for i in redword)

        def _latex_(self):
            """
            EXAMPLES::

                sage: W = WeylGroup(['A',3], prefix='s', implementation='permutation')
                sage: [s1,s2,s3] = W.simple_reflections()
                sage: s1*s2
                s1*s2
                sage: W = WeylGroup(['A',3], implementation='permutation')
                sage: [s1,s2,s3] = W.simple_reflections()
                sage: s1*s2
                (1,10,2)(3,5,6)(4,8,7)(9,11,12)
            """
            if self.parent()._prefix is None:
                return RealReflectionGroupElement._repr_(self)
            redword = self.reduced_word()
            if not redword:
                return "1"
            return "".join("%s_{%d}" % (self.parent()._prefix, i)
                           for i in redword)
